"""
vmec tools revolving around the Run class
"""
import copy
import json
import os
import numpy as np
import time
import logging
import pathlib
import rna
import sympy
import scipy
from symfit import Variable, Fit, Model, parameters, variables
import w7x

input_transcoding = w7x.transcodings.vmec_input.get_transcoding()
threed1_transcoding = w7x.transcodings.vmec_threed1.get_transcoding()


class TooLongError(ValueError):
    pass


THISDIR = pathlib.Path(__file__).resolve()


def fit(fun, x, y):
    """
    Fit a sympy function fun with data x, y
    """
    f = Variable("f")
    model = Model({f: fun})
    fit = Fit(model, x, y)
    res = fit.execute()
    fun = model[f]
    for p in res.params:
        fun = fun.subs(p, res.params.get_value(p))
    return fun


def pad(seq, target_length, padding=None):
    """
    Extend the sequence seq with padding (default: None) so as to make
    its length up to target_length. Return copy of seq. If seq is already
    longer than target_length, raise TooLongError.

    Examples:
        >>> from w7x.vmec import pad
        >>> pad([], 5, 1)
        [1, 1, 1, 1, 1]
        >>> pad([1, 2, 3], 7)
        [1, 2, 3, None, None, None, None]
        >>> pad([1, 2, 3], 2)  # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
          ...
        TooLongError: sequence too long (3) for target length 2

    """
    seq = seq[:]
    length = len(seq)
    if length > target_length:
        raise TooLongError(
            "sequence too long ({}) for target length {}".format(length, target_length)
        )
    seq.extend([padding] * (target_length - length))
    return seq


def make_radial_grid(points):
    """Make radial grid based on the number of raidal grid points.

    Args:
        points (List): numGridPointsRadial vmec parameter.

    Returns:
        numpy.array: array of radial grid points.
    """

    try:
        num_radial_points = points[-1]
    except IndexError:
        num_radial_points = points
    if num_radial_points > 20:
        num_radial_points = 20
    return np.linspace(0.0, 1.0, num_radial_points)


class Points3D(w7x.flt.Points3D):
    ws_server = w7x.Server.addr_vmec_server


class VMECBase(w7x.core.Base):
    ws_server = w7x.Server.addr_vmec_server


class Profile(VMECBase):
    prop_defaults = {"ProfileType": None, "coefficients": []}

    ws_class = "Profile"

    def __call__(self, value):
        raise NotImplementedError("A Profile must implement the __call__" "function")

    def normalized(self):
        raise NotImplementedError("A Profile must implement the normalized" "function")

    def scale(self, scale):
        self.coefficients = [c * scale for c in self.coefficients]

    def deviation(self, other):
        """
        Returns the normalized deviation of the profiles.
        """
        if self.ProfileType != other.ProfileType:
            return np.inf
        j = max(len(self.coefficients), len(other.coefficients))
        dev = np.sqrt(
            float(
                sum(
                    map(
                        lambda x: (x[0] - x[1]) ** 2,
                        zip(
                            pad(self.normalized(), j, padding=0),
                            pad(other.normalized(), j, padding=0),
                        ),
                    )
                )
            )
        )
        return dev

    def plot(self, **kwargs):
        raise NotImplementedError("A Profile must implement the plot function")


class PowerSeries(Profile):
    """
    Examples:
        >>> import w7x
        >>> default_pressure_profile = w7x.vmec.PowerSeries()

    """

    prop_defaults = dict(Profile.prop_defaults)
    prop_defaults["ProfileType"] = "power_series"
    prop_defaults["coefficients"] = [1e-6, -1e-6]

    @property
    def norm(self):
        """
        the first coefficient
        """
        if len(self.coefficients) == 0:
            return 0
        return self.coefficients[0]

    @norm.setter
    def norm(self, norm):
        self.coefficients = [norm * x for x in self.normalized()]

    def normalized(self):
        """
        Returns:
            normalized coefficients
        """
        if len(self.coefficients) == 0:
            return []
        if self.norm == 0:
            if any([c != 0 for c in self.coefficients]):
                raise ValueError("Norm is 0")
            return [0.0 for x in self.coefficients]
        return [x * 1.0 / self.norm for x in self.coefficients]

    def set_value(self, radial, value):
        """Constrain profile to satisfy (radial, value) tuple.

        Args:
            radial (float): radial point where to constrain.
            value (float): value to satisfy.
        """
        if len(self.coefficients) == 1:
            self.coefficients = [value]
        elif len(self.coefficients) == 2:
            self.coefficients[1] = (value - self.norm) / radial
        else:
            raise RuntimeError(
                "Profile with {n} coefficients is not supported".format(
                    n=len(self.coefficients)
                )
            )

    def __call__(self, value):
        poly = np.polynomial.polynomial.Polynomial(self.coefficients)
        return poly.__call__(value)

    def plot(self, **kwargs):
        rna.plotting.plot_function(self, **kwargs)


class CubicSpline(Profile):
    """
    Examples:
        >>> import w7x
        >>> default_pressure_profile = w7x.vmec.CubicSpline()

    """

    prop_defaults = dict(Profile.prop_defaults)
    prop_defaults["ProfileType"] = "cubic_spline"
    prop_defaults["coefficients"] = [1e-6, 1e-6, 5e-7, 0]

    @property
    def norm(self):
        """Retrieve the first coefficient.

        Returns:
            int: first coefficients
        """
        if len(self.coefficients) == 0:
            return 0
        return self.coefficients[0]

    @norm.setter
    def norm(self, norm):
        self.coefficients = [norm * x for x in self.normalized()]

    def normalized(self):
        """Compute normalized coefficients.

        Returns:
            list: normalized coefficients
        """
        if len(self.coefficients) == 0:
            return []
        if self.norm == 0:
            if any([c != 0 for c in self.coefficients]):
                raise ValueError("Norm is 0")
            return [0.0 for x in self.coefficients]
        return [x * 1.0 / self.norm for x in self.coefficients]

    def __call__(self, value):
        radial_points = np.linspace(0, 1, len(self.coefficients))
        spline = scipy.interpolate.CubicSpline(radial_points, self.coefficients)
        return spline.__call__(value)

    def plot(self, **kwargs):
        rna.plotting.plot_function(self, **kwargs)


class TwoPower(Profile):
    """
    Examples:
        >>> import w7x
        >>> default_pressure_profile = w7x.vmec.TwoPower()

    """

    prop_defaults = dict(Profile.prop_defaults)
    prop_defaults["ProfileType"] = "two_power"
    prop_defaults["coefficients"] = [1e-6, 1.0, 1.0]

    @property
    def norm(self):
        """
        the first coefficient
        """
        if len(self.coefficients) == 0:
            return 0
        return self.coefficients[0]

    @norm.setter
    def norm(self, norm):
        self.coefficients = [norm * x for x in self.normalized()]

    def normalized(self):
        """
        Returns:
            normalized coefficients
        """
        if len(self.coefficients) == 0:
            return []
        if self.norm == 0:
            if any([c != 0 for c in self.coefficients]):
                raise ValueError("Norm is 0")
            return [0.0 for x in self.coefficients]
        return [x * 1.0 / self.norm for x in self.coefficients]

    def __call__(self, value):
        return (
            self.coefficients[0]
            * (1 - value ** self.coefficients[1]) ** self.coefficients[2]
        )

    def plot(self, **kwargs):
        rna.plotting.plot_function(self, **kwargs)


class SumAtan(Profile):
    """
    Examples:
        >>> import w7x
        >>> default_current_profile = w7x.vmec.SumAtan()

    """

    prop_defaults = dict(Profile.prop_defaults)
    prop_defaults["ProfileType"] = "sum_atan"
    prop_defaults["coefficients"] = [0, 1.0, 1.0, 1.0, 1.0]

    #  TODO(@amerlo): State what profile norm means here
    @property
    def norm(self):
        """
        the first coefficient
        """
        if len(self.coefficients) == 0:
            return 0
        return self.coefficients[0]

    @norm.setter
    def norm(self, norm):
        self.coefficients = [norm * x for x in self.normalized()]

    def normalized(self):
        """
        Returns:
            normalized coefficients
        """
        if len(self.coefficients) == 0:
            return []
        if self.norm == 0:
            if any([c != 0 for c in self.coefficients]):
                raise ValueError("Norm is 0")
            return [0.0 for x in self.coefficients]
        return [x * 1.0 / self.norm for x in self.coefficients]

    def __call__(self, value):
        return self.coefficients[0] + 2 / np.pi * self.coefficients[1] * np.arctan(
            np.divide(
                self.coefficients[2] * value ** self.coefficients[3],
                (1 - value) ** self.coefficients[4],
            )
        )

    def plot(self, **kwargs):
        rna.plotting.plot_function(self, **kwargs)


def create_profile(*args, **kwargs):
    """
    Factory method for profile creation
    """
    log = logging.getLogger()
    profileType = kwargs["ProfileType"]
    if profileType == "power_series":
        return PowerSeries(*args, **kwargs)
    elif profileType == "cubic_spline":
        return CubicSpline(*args, **kwargs)
    elif profileType == "two_power":
        return TwoPower(*args, **kwargs)
    elif profileType == "sum_atan":
        return SumAtan(*args, **kwargs)
    else:
        log.error("No Profile subclass matches type {profileType}.".format(**locals()))
        return Profile(*args, **kwargs)


class FourierCoefficients(VMECBase):
    ws_class = "FourierCoefficients"
    prop_defaults = {
        "coefficients": None,
        "poloidalModeNumbers": None,  # m
        "toroidalModeNumbers": None,  # n
        "numRadialPoints": 0,
    }

    def __init__(self, *args, **kwargs):
        set_null = kwargs.pop("set_null", True)
        super(FourierCoefficients, self).__init__(*args, **kwargs)
        # check coefficients for negative m=0,n<0 to be 0
        if set_null:
            for i, n in enumerate(self.toroidalModeNumbers):
                # m = 0
                if n < 0:
                    log = logging.getLogger()
                    log.debug(
                        "For m=0, n<0 coefficients may not be != 0. "
                        "Set them to be 0."
                    )
                    if self.coefficients[i] != 0.0:
                        self.coefficients[i] = 0.0
                else:
                    break
        nCoeffs = len(self.coefficients)
        nTor = len(self.toroidalModeNumbers)
        mPol = len(self.poloidalModeNumbers)
        nRad = self.numRadialPoints
        if not nCoeffs == nTor * mPol * nRad:
            raise ValueError(
                "len(coeffs)({nCoeffs}) != nTor({nTor}) *"
                " mPol({mPol}) *"
                " nRad({nRad})".format(**locals())
            )

    def __call__(self, radial, phi, theta, component):
        values = []

        if component not in ["cos", "sin"]:
            raise ValueError("Type of Fourier Coefficients not supported")

        if radial >= self.numRadialPoints:
            raise ValueError(
                "Radial location {r} is outside coefficients boundary".format(r=radial)
            )

        i = radial * len(self.poloidalModeNumbers) * len(self.toroidalModeNumbers)
        for m in self.poloidalModeNumbers:
            for n in self.toroidalModeNumbers:
                if component == "cos":
                    part = self.coefficients[i] * np.cos(m * theta - 5 * n * phi)
                else:
                    part = self.coefficients[i] * np.sin(m * theta - 5 * n * phi)
                values.append(part)
                i += 1

        return sum(values)

    def plot(self, *args, **kwargs):
        """Plot FourierCoefficients."""
        w7x.plotting.vmec.plot_fourier(self, *args, **kwargs)

    @property
    def shape(self):
        """Return shape of FourierCoefficients."""

        return (
            self.numRadialPoints,
            len(self.poloidalModeNumbers),
            len(self.toroidalModeNumbers),
        )

    def reshape(self, shape, keepdims=False):
        """Reshape FourierCoefficients to `shape` structure.

        The output coefficients will have poloidal and toroidal modes based
        on the shape parameter.

        Args:
            shape (tuple): (nRad, mPol, nTor) tuple.
            keepdims (bool): if true, output shape is the same as object.

        Returns:
            w7x.vmec.FourierCoefficients: output coefficients.
        """

        if self.shape == shape:
            return self

        nRad = np.linspace(0, self.numRadialPoints - 1, shape[0], dtype=int)
        mPol = range(shape[1])
        nTor = range(-int((shape[2] - 1) / 2), int((shape[2] - 1) / 2 + 1))

        indices = np.arange(len(self.coefficients))
        indices = indices.reshape(self.shape)
        indices = [
            indices[r, m, n + self.toroidalModeNumbers[-1]]
            for r in nRad
            for m in mPol
            for n in nTor
        ]

        if keepdims:
            coeffs = np.zeros(len(self.coefficients))
            coeffs[indices] = self.coefficients[indices]

            return FourierCoefficients(
                coefficients=coeffs,
                toroidalModeNumbers=self.toroidalModeNumbers,
                poloidalModeNumbers=self.poloidalModeNumbers,
                numRadialPoints=self.numRadialPoints,
            )

        coeffs = np.zeros(shape[0] * shape[1] * shape[2])
        coeffs = self.coefficients[indices]

        return FourierCoefficients(
            coefficients=coeffs,
            toroidalModeNumbers=nTor,
            poloidalModeNumbers=mPol,
            numRadialPoints=nRad,
        )

    @classmethod
    def get_indices(cls, poloidal_modes, toroidal_modes, modes=12):
        """Compute flatted coefficient indeces from poloidal and toroidal modes.

        Args:
            poloidal_modes (list): list of poloidal modes.
            toroidal_modes (list): list of toroidal modes.
            modes (int): number of fourier modes.

        Returns:
            list: list of flatted coefficients indeces as VMEC reppresentation.
        """

        coeffs = np.arange(modes ** 2 * 2)
        coeffs = np.insert(coeffs, 0, [0] * modes)
        coeffs = coeffs.reshape(modes, modes * 2 + 1)

        indeces = [
            coeffs[m, n]
            for m in poloidal_modes
            for n in [mode + modes for mode in toroidal_modes]
        ]

        return indeces

    @classmethod
    def from_array(cls, coefficients):
        """Return FourierCoefficients from array.

        We assume that coefficients have the same number of poloidal and
        toroidal modes.

        Args:
            coefficients (numpy.ndarray): input array.

        Returns:
            w7x.vmec.FourierCoefficients: output fourier coefficients.
        """

        radial_points = coefficients.shape[0]
        poloidal_modes = range(int(np.sqrt(coefficients.shape[1] / 2)))
        toroidal_modes = range(-poloidal_modes[-1] - 1, poloidal_modes[-1] + 2)

        #  Extend coefficients for m=0, n<0
        coefficients = np.apply_along_axis(
            lambda x: np.insert(x, 0, [0] * len(poloidal_modes)), 1, coefficients
        )

        return cls(
            coefficients=coefficients.flatten(),
            poloidalModeNumbers=poloidal_modes,
            toroidalModeNumbers=toroidal_modes,
            numRadialPoints=radial_points,
        )


class SurfaceCoefficients(VMECBase):
    ws_class = "surfaceCoefficients"
    prop_defaults = {"RCos": None, "ZSin": None, "RSin": None, "ZCos": None}

    def __init__(self, *args, **kwargs):
        super(SurfaceCoefficients, self).__init__(*args, **kwargs)
        """
        The wsDoku says you need to define hybrid but haukes code shows cylindrical.
        I will take default hybrid grid.
        It appears to not have any effect to change between the two.
        """
        if self.RCos is None:
            self.RCos = FourierCoefficients()
        if self.ZSin is None:
            self.ZSin = FourierCoefficients()

        if self.RCos.numRadialPoints != self.ZSin.numRadialPoints:
            raise ValueError("Coefficients radial resolutions are not the same")

    @classmethod
    def default_axis(cls):
        r_cos = [5.54263, 0.184047]
        z_sin = [0.0000, 0.157481]

        # Fourier coefficients for the boundary
        coeffs_r_cos = FourierCoefficients(
            coefficients=r_cos,
            poloidalModeNumbers=[0],
            toroidalModeNumbers=[0, 1],
            numRadialPoints=1,
        )
        coeffs_z_sin = FourierCoefficients(
            coefficients=z_sin,
            poloidalModeNumbers=[0],
            toroidalModeNumbers=[0, 1],
            numRadialPoints=1,
        )

        return cls(RCos=coeffs_r_cos, ZSin=coeffs_z_sin)

    @classmethod
    def default_boundary(cls):
        # m      0    0   0       1      1    1
        # n      -1   0   1      -1      0    1
        r_cos = [0.0, 5.5, 0.151, 0.039, 0.26, -0.085]
        z_sin = [0.0, 0.0, -0.154, 0.040, 0.45, 0.180]

        # Fourier coefficients for the boundary
        coeffs_r_cos = FourierCoefficients(
            coefficients=r_cos,
            poloidalModeNumbers=[0, 1],
            toroidalModeNumbers=[-1, 0, 1],
            numRadialPoints=1,
        )
        coeffs_z_sin = FourierCoefficients(
            coefficients=z_sin,
            poloidalModeNumbers=[0, 1],
            toroidalModeNumbers=[-1, 0, 1],
            numRadialPoints=1,
        )

        return cls(RCos=coeffs_r_cos, ZSin=coeffs_z_sin)

    def __call__(self, radial, phi, theta):
        partsR = []
        partsZ = []

        if radial >= self.RCos.numRadialPoints:
            raise ValueError(
                "Radial location {r} is outside coefficients boundary".format(r=radial)
            )

        i = (
            radial
            * len(self.RCos.poloidalModeNumbers)
            * len(self.RCos.toroidalModeNumbers)
        )
        for m in self.RCos.poloidalModeNumbers:
            for n in self.RCos.toroidalModeNumbers:
                part = self.RCos.coefficients[i] * np.cos(m * theta - 5 * n * phi)
                partsR.append(part)
                i += 1

        j = (
            radial
            * len(self.RCos.poloidalModeNumbers)
            * len(self.RCos.toroidalModeNumbers)
        )
        for m in self.ZSin.poloidalModeNumbers:
            for n in self.ZSin.toroidalModeNumbers:
                part = self.ZSin.coefficients[j] * np.sin(m * theta - 5 * n * phi)
                partsZ.append(part)
                j += 1
        return sum(partsR), sum(partsZ)

    def plot(self, *args, **kwargs):
        """Plot SurfaceCoefficients."""
        w7x.plotting.vmec.plot_surface(self, *args, **kwargs)


def vmec_id_exists(vmec_id):
    vmec_server = w7x.get_server(w7x.Server.addr_vmec_server)
    return vmec_server.service.vmecIdentifierExists(vmec_id)


def volume_lcfs(vmec_id):
    vmec_server = w7x.get_server(w7x.Server.addr_vmec_server)
    return vmec_server.service.getVolumeLCFS(vmec_id)


def was_successful(vmec_id):
    vmec_server = w7x.get_server(w7x.Server.addr_vmec_server)
    try:
        success = vmec_server.service.wasSuccessful(vmec_id)
    except Exception:
        log = logging.getLogger()
        success = False
        log.error("@vmec_id '{vmec_id}': {err}".format(**locals()))
    return success


def existing_vmec_ids():
    vmec_server = w7x.get_server(w7x.Server.addr_vmec_server)
    res = vmec_server.service.listIdentifiers()
    return res.ReferenceShortIds + res.VmecIds


def getBAxis(vmec_id, phi=0.0):
    """
    Args:
        vmec_id (str): vmec identifier
        phi (float): phi in radian
    Returns:
        float: Bax(phi) - magnetic field magnitude on the magnetic
            axis at phi = <phi>
    """
    vmec_server = w7x.get_server(w7x.Server.addr_vmec_server)
    points = Points3D(vmec_server.service.getMagneticAxis(vmec_id, phi))
    res = vmec_server.service.magneticField(vmec_id, points.as_input())
    B = Points3D(res)
    return np.linalg.norm(B)


def get_initial_from_w7xfp(coil_currents, pressure, current):
    """Retrieve initial magneticAxis and boundary from W7-X FP.

    Args:
        coil_currents: list of coil currents in 'A'.
        pressure: pressure profile.
        current: current profile.

    Returns:
        magneticAxis, boundary: initial condition for VMEC run.
    """

    vmec_server = w7x.get_server(w7x.Server.addr_w7xfp_server)

    knots = vmec_server.service.getProfilesNumberOfKnots()
    radial_grid = np.linspace(0.0, 1.0, knots)

    pressure = pressure(radial_grid)
    jtor = current(radial_grid)
    plasma_radius = 0.52
    Rmn = vmec_server.service.getFourierCoefficientsRmn(
        pressure, jtor, plasma_radius, coil_currents
    )
    Zmn = vmec_server.service.getFourierCoefficientsZmn(
        pressure, jtor, plasma_radius, coil_currents
    )
    mPol = len(Rmn.poloidalModeNumbers)
    nTor = len(Rmn.toroidalModeNumbers)
    #  Use only non null coefficients for the axis
    magneticAxis_RCos = w7x.vmec.FourierCoefficients(
        coefficients=Rmn.coefficients[::knots][mPol - 1 : mPol + 1],
        poloidalModeNumbers=[0],
        toroidalModeNumbers=[0, 1],
        numRadialPoints=1,
    )
    magneticAxis_ZSin = w7x.vmec.FourierCoefficients(
        coefficients=Zmn.coefficients[::knots][mPol - 1 : mPol + 1],
        poloidalModeNumbers=[0],
        toroidalModeNumbers=[0, 1],
        numRadialPoints=1,
    )
    magneticAxis = w7x.vmec.SurfaceCoefficients(
        RCos=magneticAxis_RCos, ZSin=magneticAxis_ZSin
    )

    #  Use only 0,1 coefficients
    boundary_RCos = w7x.vmec.FourierCoefficients(
        coefficients=Rmn.coefficients[knots - 1 :: knots][mPol - 2 : mPol + 1]
        + Rmn.coefficients[knots - 1 :: knots][nTor + mPol - 2 : nTor + mPol + 1],
        poloidalModeNumbers=[0, 1],
        toroidalModeNumbers=[-1, 0, 1],
        numRadialPoints=1,
    )
    boundary_ZSin = w7x.vmec.FourierCoefficients(
        coefficients=Zmn.coefficients[knots - 1 :: knots][mPol - 2 : mPol + 1]
        + Zmn.coefficients[knots - 1 :: knots][nTor + mPol - 2 : nTor + mPol + 1],
        poloidalModeNumbers=[0, 1],
        toroidalModeNumbers=[-1, 0, 1],
        numRadialPoints=1,
    )
    boundary = w7x.vmec.SurfaceCoefficients(RCos=boundary_RCos, ZSin=boundary_ZSin)

    return (magneticAxis, boundary)


def vmec_input(**kwargs):
    """
    Args:
        **kwargs
            Necessary:
                magnetic_config (MagneticConfig)
                pressure_profile (Profile)
                current_profile (Profile)
                magneticAxis (FourierCoefficients)
                boundary (FourierCoefficients)
            Obligatory:
                maxIterationsPerSequence (int): -> NITER
                maxToroidalMagneticFlux (float): -> PHIEDGE

    """

    magnetic_config = kwargs.pop("magnetic_config")
    pressure_profile = kwargs.pop("pressure_profile")
    current_profile = kwargs.pop("current_profile")
    magneticAxis = kwargs.pop("magneticAxis")
    boundary = kwargs.pop("boundary")
    iotaProfile = kwargs.pop("iotaProfile", None)
    pressureScale = kwargs.pop("pressureScale", w7x.Defaults.VMEC.pressureScale)
    totalToroidalCurrent = kwargs.pop(
        "totalToroidalCurrent", w7x.Defaults.VMEC.totalToroidalCurrent
    )
    maxIterationsPerSequence = kwargs.pop(
        "maxIterationsPerSequence", w7x.Defaults.VMEC.maxIterationsPerSequence
    )
    maxToroidalMagneticFlux = kwargs.pop(
        "maxToroidalMagneticFlux", w7x.Defaults.VMEC.maxToroidalMagneticFlux
    )
    timeStep = kwargs.pop("timeStep", w7x.Defaults.VMEC.timeStep)
    numGridPointsRadial = kwargs.pop(
        "numGridPointsRadial", w7x.Defaults.VMEC.numGridPointsRadial
    )
    forceToleranceLevels = kwargs.pop(
        "forceToleranceLevels", w7x.Defaults.VMEC.forceToleranceLevels
    )
    numModesPoloidal = kwargs.pop("numModesPoloidal", 12)
    numModesToroidal = kwargs.pop("numModesToroidal", 12)
    getInitialGuess = kwargs.pop("getInitialGuess", False)

    #  See VMEC Input Namelist documentation on STELLOPT repository
    numGridPointsPoloidal = kwargs.pop(
        "numGridPointsPoloidal", 2 * numModesPoloidal + 6
    )

    vmec_server = w7x.get_server(w7x.Server.addr_vmec_server)
    input_data = vmec_server.types.VmecInput(True)

    input_data.mgridFile = "mgrid_w7x_nv36_hires.nc"  # 'mgrid_w7x_nv36.nc'
    input_data.coilCurrents = magnetic_config.coil_currents("A")

    input_data.pressureProfile = pressure_profile.as_input()
    input_data.toroidalCurrentProfile = current_profile.as_input()
    if iotaProfile:
        input_data.iotaProfile = iotaProfile.as_input()

    #  Retrieve magneticAxis and boundary from W7-X function parametrization
    #  http://webservices.ipp-hgw.mpg.de/docs/w7xfp.html
    if getInitialGuess:
        (magneticAxis, boundary) = get_initial_from_w7xfp(
            coil_currents=magnetic_config.coil_currents("A"),
            pressure=pressure_profile,
            current=current_profile,
        )

    input_data.freeBoundary = True
    input_data.intervalFullVacuumCalculation = 6
    input_data.numFieldPeriods = 5
    input_data.numModesPoloidal = numModesPoloidal
    input_data.numModesToroidal = numModesToroidal
    input_data.numGridPointsPoloidal = numGridPointsPoloidal
    input_data.numGridPointsToroidal = 36
    input_data.numGridPointsRadial = numGridPointsRadial
    input_data.forceToleranceLevels = forceToleranceLevels
    input_data.pressureScale = pressureScale
    input_data.totalToroidalCurrent = totalToroidalCurrent
    input_data.timeStep = timeStep
    input_data.tcon0 = 2.0
    input_data.maxIterationsPerSequence = maxIterationsPerSequence
    input_data.intervalConvergenceOutput = 100
    input_data.maxToroidalMagneticFlux = maxToroidalMagneticFlux
    input_data.gamma = 0
    input_data.magneticAxis = magneticAxis.as_input()
    input_data.boundary = boundary.as_input()
    return input_data


def find_run(
    currents_rw,
    rel_pressure_profile=[1, -1],
    rel_current_profile=[],
    delta=1e-2,
    min_beta=None,
    max_beta=None,
    min_b_axis=None,
    max_b_axis=None,
    minP0=None,
    maxP0=None,
    idContains=None,
    ready=False,
):
    log = logging.getLogger()
    vmec_ids = existing_vmec_ids()
    pressure_profile = (
        PowerSeries(coefficients=rel_pressure_profile)
        if rel_pressure_profile is not None
        else None
    )
    current_profile = (
        PowerSeries(coefficients=rel_current_profile)
        if rel_current_profile is not None
        else None
    )
    for vmec_id in vmec_ids:
        log.debug("vmec_id {vmec_id}:".format(**locals()))
        if idContains is not None:
            if not all([word in vmec_id for word in idContains]):
                log.debug(
                    "\t\t\t...vmec_id does not contain one of "
                    "idContains({idContains})".format(**locals())
                )
                continue

        run = Run(vmec_id)
        """coilCurrents"""
        try:
            run_currents = run.magnetic_config.coil_currents("rw")
        except Exception:
            log.debug("\t\t\t...error when getting magneitcConfig")
            continue
        if not run.was_successful():
            if ready:
                log.debug("\t\t\t...not successful")
                continue
            elif run.is_ready():
                log.debug("\t\t\t...not successful")
                continue
        dev = np.sqrt(
            sum(map(lambda x: abs((x[0] - x[1]) ** 2), zip(run_currents, currents_rw)))
        )

        if dev > delta:
            log.debug(
                "\t\t\t...differing in magnetic_config by {dev}".format(**locals())
            )
            continue

        """Pressure"""
        if pressure_profile is not None:
            if not isinstance(run.pressure_profile, PowerSeries):
                log.debug("\t\t\t...PressureProfile is not PowerSeries")
                continue
            p0 = run.pressure_profile.norm
            if minP0 is not None:
                if p0 < minP0:
                    log.debug("\t\t\t...p0({p0}) < minP0({minP0})".format(**locals()))
                    continue
            if maxP0 is not None:
                if p0 > maxP0:
                    log.debug("\t\t\t...p0({p0}) > maxP0({maxP0})".format(**locals()))
                    continue
            devPP = pressure_profile.deviation(run.pressure_profile)
            if devPP > delta:
                log.debug(
                    "\t\t\t...differing in relative pressure_profile by"
                    " {devPP}".format(**locals())
                )
                continue

        """Current"""
        if current_profile is not None:
            if not isinstance(run.current_profile, PowerSeries):
                log.debug("\t\t\t...CurrentProfile is not PowerSeries")
                continue
            devCP = current_profile.deviation(run.current_profile)
            if devCP > delta:
                log.debug(
                    "\t\t\t...differing in relative current_profile by"
                    " {devCP}".format(**locals())
                )
                continue

        """Beta"""
        if min_beta is not None or max_beta is not None:
            beta = run.beta()
        if min_beta is not None:
            if not min_beta < beta:
                log.debug(
                    "\t\t\t...min_beta({min_beta}) > beta({beta})".format(**locals())
                )
                continue
        if max_beta is not None:
            if not beta < max_beta:
                log.debug(
                    "\t\t\t...max_beta({max_beta}) < beta({beta})".format(**locals())
                )
                continue

        """BAxis"""
        if min_b_axis is not None or max_b_axis is not None:
            b_axis = run.getBAxis()
        if min_b_axis is not None:
            if not min_b_axis < b_axis:
                log.debug(
                    "\t\t\t...min_b_axis({min_b_axis}) > b_axis({b_axis})".format(
                        **locals()
                    )
                )
                continue
        if max_b_axis is not None:
            if not b_axis < max_beta:
                log.debug(
                    "\t\t\t...max_b_axis({max_b_axis}) < b_axis({b_axis})".format(
                        **locals()
                    )
                )
                continue

        return run


class Run(object):
    RUNARGS = [
        "magnetic_config",
        "pressure_profile",
        "current_profile",
        "magneticAxis",
        "boundary",
    ]

    def __init__(self, *args, **run_kwargs):
        """
        Examples:
            >>> import w7x

            Picking up an old run
            >>> run = w7x.vmec.Run("w7x_ref_9")
            >>> run.vmec_id
            'w7x_ref_9'

            or
            >>> run = w7x.vmec.Run(run)

            A run has the attributes:
            >>> run.vmec_id
            'w7x_ref_9'

            auto vmec_id generation
            >>> vmec_id = w7x.vmec.Run(magnetic_config=w7x.flt.MagneticConfig.default(),
            ...                        pressure_profile=w7x.vmec.PowerSeries(),
            ...                        current_profile=w7x.vmec.PowerSeries()).vmec_id
            >>> vmec_id.startswith('w7x')
            True

            magnetic_config
            >>> run.magnetic_config.coil_currents('A')  # noqa
            [15000.0, 15000.0, 15000.0, 15000.0, 15000.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]

            plasma beta
            # TODO(@dboe): I fear the vmec output definition changed again
            # >>> run.beta()
            # 0.01984

        """
        self._iteration = run_kwargs.pop("iteration", 0)
        self.parent = run_kwargs.pop("parent", None)
        self._folder = run_kwargs.pop("folder", "~/tmp/VMEC/")
        self._run_kwargs = {}
        self.run_kwargs = run_kwargs
        self._vmec_id = None
        if len(args) == 0:
            vmec_id = self.build_vmec_id()
        elif len(args) == 1:
            vmec_id = args[0]
        else:
            raise ValueError("Too many arguments.")
        self.vmec_id = vmec_id
        self._parsed = []
        self._backend = run_kwargs.pop("backend", "web_service")

    @classmethod
    def vacuum(cls, magnetic_config=None, **kwargs):
        if magnetic_config is None:
            raise ValueError("magnetic_config missing")
        kwargs["magnetic_config"] = magnetic_config
        kwargs["pressure_profile"] = PowerSeries()
        kwargs["current_profile"] = PowerSeries()
        if "parent" not in kwargs:
            kwargs["magneticAxis"] = SurfaceCoefficients.default_axis()
            kwargs["boundary"] = SurfaceCoefficients.default_boundary()

        return cls(**kwargs)

    def __str__(self):
        string = (
            "{self.__class__} instance\n"
            "\n"
            "\t\t\tvmec_id: {self.vmec_id}\n"
            "\n".format(**locals())
        )

        """ add the status """
        if not vmec_id_exists(self.vmec_id):
            status = "initializing"
        elif self.was_successful():
            status = "successful"
        else:
            status = "not successful"
        string += "Status: {status}\n".format(**locals())

        """ add the input parameters """
        if self.parent is None:
            partent_vmec_id = "None"
        elif not vmec_id_exists(self.parent.vmec_id):
            partent_vmec_id = "unknown"
        else:
            partent_vmec_id = self.parent.vmec_id
        string += (
            "Parameters:\n"
            "\tparent: {partent_vmec_id}\n"
            "\tcoil currents (winding currents in A): {coilCurrents}\n".format(
                coilCurrents=self.magnetic_config.coil_currents("A"), **locals()
            )
        )
        string += (
            "\tpressure_profile:\n"
            "\t\tType: {self.pressure_profile.ProfileType}\n"
            "\t\tcoefficients: {self.pressure_profile.coefficients}\n".format(
                **locals()
            )
        )
        string += (
            "\tcurrentProfile:\n"
            "\t\tType: {self.current_profile.ProfileType}\n"
            "\t\tcoefficients: {self.current_profile.coefficients}\n".format(**locals())
        )
        restKwargs = copy.deepcopy(self.run_kwargs)
        for key in restKwargs:
            if key not in self.RUNARGS:
                string += "{key}: {value}\n".format(key=key, value=restKwargs[key])

        """ add the major results """
        if self.was_successful():
            string += (
                "\n"
                "Results:\n"
                "\tB_Axis(phi=0): {b_axis}\n"
                "\tbeta: {beta}\n"
                "\tvolume LCFS (m^3): {volume}\n"
                "\tforce: {force}\n".format(
                    b_axis=self.getBAxis(),
                    beta=self.beta(),
                    volume=self.volume_lcfs(),
                    force=self.force(),
                    **locals()
                )
            )
        return string

    @property
    def vmec_id(self):
        return self._vmec_id

    @vmec_id.setter
    def vmec_id(self, vmec_id):
        if isinstance(vmec_id, Run):
            vmec_id = vmec_id.vmec_id
        self._vmec_id = vmec_id

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, parent):
        if isinstance(parent, str):
            parent = Run(parent)
        self._parent = parent

    @property
    def run_kwargs(self):
        """
        (dict): kwargs to pass to the vmec service
        """
        return self._run_kwargs

    @run_kwargs.setter
    def run_kwargs(self, run_kwargs):
        self._run_kwargs = run_kwargs

    @property
    def magnetic_config(self):
        """
        (w7x.flt.MagneticConfig)
        """
        magnetic_config = self.run_kwargs.get("magnetic_config", None)
        if magnetic_config is not None:
            return magnetic_config
        if self.parent is not None:
            return self.parent.magnetic_config
        if self.exists():
            vmec_server = w7x.get_server(w7x.Server.addr_vmec_server)
            currents = np.array(vmec_server.service.getCoilCurrents(self.vmec_id))
            magnetic_config = w7x.MagneticConfig.from_currents(*currents, unit="A")
            return magnetic_config
        raise ValueError("No magnetic_config found.")

    @magnetic_config.setter
    def magnetic_config(self, config):
        self.run_kwargs["magnetic_config"] = config

    @property
    def pressure_profile(self):
        pressure_profile = self.run_kwargs.get("pressure_profile", None)
        if pressure_profile is not None:
            return pressure_profile
        if self.parent is not None:
            return self.parent.pressure_profile
        if self.exists():
            self._parse_input()
            pressure_profile = create_profile(
                ProfileType=self._input["pressureProfileProfileType"],
                coefficients=self._input["pressureProfileCoefficients"],
            )
            return pressure_profile
        raise ValueError("No pressure_profile found.")

    @pressure_profile.setter
    def pressure_profile(self, profile):
        if profile.norm == 0:
            profile = PowerSeries(coefficients=[1e-6, -1e-6])  # very slight gradient
        self.run_kwargs["pressure_profile"] = profile

    @property
    def current_profile(self):
        current_profile = self.run_kwargs.get("current_profile", None)
        if current_profile is not None:
            return current_profile
        if self.parent is not None:
            return self.parent.current_profile
        if self.exists():
            self._parse_input()
            current_profile = create_profile(
                ProfileType=self._input["currentProfileProfileType"],
                coefficients=self._input["currentProfileCoefficients"],
            )
            return current_profile
        raise ValueError("No current_profile found.")

    @current_profile.setter
    def current_profile(self, profile):
        self.run_kwargs["current_profile"] = profile

    @property
    def magneticAxis(self):
        magneticAxis = self.run_kwargs.get("magneticAxis", None)
        if magneticAxis is not None:
            return magneticAxis
        if self.parent is not None:
            return self.parent.magneticAxis
        if self.exists():
            self._parse_threed1()
            rcos = FourierCoefficients(
                coefficients=self._threed1["rac"][:2],
                poloidalModeNumbers=[0],
                toroidalModeNumbers=[0, 1],
                numRadialPoints=1,
            )
            zsin = FourierCoefficients(
                coefficients=self._threed1["zas"][:2],
                poloidalModeNumbers=[0],
                toroidalModeNumbers=[0, 1],
                numRadialPoints=1,
            )
            magneticAxis = SurfaceCoefficients(RCos=rcos, ZSin=zsin)
        return magneticAxis

    @magneticAxis.setter
    def magneticAxis(self, surfaceCoefficients):
        self.run_kwargs["magneticAxis"] = surfaceCoefficients

    @property
    def boundary(self):
        """
        (SurfaceCoefficients): fourrier description of the vmec boundary
        """
        boundary = self.run_kwargs.get("boundary", None)
        if boundary is not None:
            return boundary
        if self.parent is not None:
            return self.parent.boundary
        if self.exists():
            # parse boundary of completed run
            self._parse_threed1()
            coeffDict = {
                (m, n): (rbc, zbs)
                for m, n, rbc, zbs in zip(
                    self._threed1["mb"],
                    self._threed1["nb"],
                    self._threed1["rbc"],
                    self._threed1["zbs"],
                )
            }

            toroidalModeNumbers = sorted(
                set([x for x in self._threed1["nb"] if abs(x) <= 6])
            )
            poloidalModeNumbers = sorted(
                set([x for x in self._threed1["mb"] if abs(x) <= 6])
            )
            rbcChoice = []
            zbsChoice = []
            for m in poloidalModeNumbers:
                for n in toroidalModeNumbers:
                    if m == 0 and n < 0:
                        rbcChoice.append(0.0)
                        zbsChoice.append(0.0)
                        """
                        the latter would be the mathematically correct version
                        Somehow the vmec service does not handle m=0,n<0 correct
                        """
                        # rbcChoice.append(coeffDict[(m, abs(n))][0])
                        # zbsChoice.append(-coeffDict[(m, abs(n))][1])
                    else:
                        rbcChoice.append(coeffDict[(m, n)][0])
                        zbsChoice.append(coeffDict[(m, n)][1])

            rcos = FourierCoefficients(
                coefficients=rbcChoice,
                poloidalModeNumbers=poloidalModeNumbers,
                toroidalModeNumbers=toroidalModeNumbers,
                numRadialPoints=1,
            )
            zsin = FourierCoefficients(
                coefficients=zbsChoice,
                poloidalModeNumbers=poloidalModeNumbers,
                toroidalModeNumbers=toroidalModeNumbers,
                numRadialPoints=1,
            )
            boundary = SurfaceCoefficients(RCos=rcos, ZSin=zsin)
        elif boundary is None:
            boundary = SurfaceCoefficients.default_boundary()
        return boundary

    @boundary.setter
    def boundary(self, surfaceCoefficients):
        self.run_kwargs["boundary"] = surfaceCoefficients

    def mutate(self, **mutationKwargs):
        """
        Change the actual run settings
        """
        if "vmec_id" in mutationKwargs:
            inst = self.__class__(mutationKwargs.pop("vmec_id"), parent=self)
        else:
            inst = self.__class__(parent=self)
        inst.run_kwargs.pop("boundary", None)
        inst.run_kwargs.pop("magneticAxis", None)
        inst.run_kwargs.update(mutationKwargs)
        return inst

    def build_vmec_id(self):
        w7x_version = w7x.__version__
        geiger_string = self.magnetic_config.geiger_string()
        pressure_norm = int(round(self.pressure_profile.norm / 1e4, 0))
        iteration = self._iteration
        vmec_id = (
            "w7x"
            "_v_{w7x_version}"
            "_id_{geiger_string}"
            "_pres_{pressure_norm:0>2d}"
            "_it_{iteration}"
        ).format(**locals())
        if vmec_id_exists(vmec_id):
            self._iteration += 1
            return self.build_vmec_id()
        return vmec_id

    def field_period(self):
        vmec_server = w7x.get_server(w7x.Server.addr_vmec_server)
        return vmec_server.service.getFieldPeriod(self.vmec_id)

    def closest_vmec_run(self, **kwargs):
        """ Find closest vmec run.

        Examples:
            >>> import w7x
            >>> run = w7x.vmec.Run('dboe_id_1000_1000_1000_1000_+0000_-0250_v_00_pres_04_it_0')

            # FIXME TODO(@amerlo): fails
            # >>> run.closest_vmec_run().vmec_id == run.vmec_id
            # True
        """

        log = logging.getLogger()
        vmec_ids = existing_vmec_ids()
        vmec_id_filter = kwargs.pop("parent", "")

        #  First find deviations in relative currents
        coil_currents = self.magnetic_config.coil_currents("rw")
        close_runs_in_currents = {}
        close_runs_in_currents[np.inf] = []

        #  Filter vmec_ids based on the prefix
        vmec_ids = [vmec_id for vmec_id in vmec_ids if vmec_id_filter in vmec_id]

        #  Cycle trough all vmec_ids and find the closest in terms of
        #  coil currents
        for vmec_id in vmec_ids:
            run = Run(vmec_id)
            try:
                run_coil_currents = run.magnetic_config.coil_currents("rw")
            except Exception:
                close_runs_in_currents[np.inf].append(vmec_id)
                continue
            if run.was_successful():
                dev = np.sqrt(
                    sum(
                        map(
                            lambda x: abs((x[0] - x[1]) ** 2),
                            zip(run_coil_currents, coil_currents),
                        )
                    )
                )
            else:
                dev = np.inf
            try:
                close_runs_in_currents[dev].append(vmec_id)
            except KeyError:
                close_runs_in_currents[dev] = [vmec_id]

        #  Get list of closest runs in term of currents
        log.info(
            "List of close vmec runs in term of coil currents  "
            "{runs}".format(
                runs=close_runs_in_currents[min(close_runs_in_currents.keys())]
            )
        )
        close_runs = list(
            map(
                lambda x: Run(x),
                close_runs_in_currents.get(min(close_runs_in_currents.keys())),
            )
        )

        if len(close_runs) == 1:
            return close_runs[0]
        elif len(close_runs) == 0:
            raise ValueError("No close runs found")

        #  Second test on deviation in pressure_profile
        close_runs_in_pressure = {}
        close_runs_in_pressure[np.inf] = []
        for run in close_runs:
            try:
                run.pressure_profile
            except Exception as err:
                log.error(str(err))
                log.warning(
                    "Probably false profile Type for vmec_id "
                    "{vmec_id}".format(vmec_id=run.vmec_id)
                )
                close_runs_in_pressure[np.inf].append(run)
                continue
            dev = self.pressure_profile.deviation(run.pressure_profile)
            try:
                close_runs_in_pressure[dev].append(run)
            except KeyError:
                close_runs_in_pressure[dev] = [run]

        min_pressure_deviation = min(close_runs_in_pressure.keys())
        if len(close_runs_in_pressure[min_pressure_deviation]) == 1:
            return close_runs_in_pressure[min_pressure_deviation]
        else:
            return close_runs_in_pressure[min_pressure_deviation][0]

    def _complete_run_kwargs(self, **kwargs):
        """Complete the necessary arguments for starting a vmec run.

        The arguments are derived from the adopting run stored in the server.
        """

        for arg in self.RUNARGS:
            if self.run_kwargs.get(arg, None) is None:
                if self.parent is None:
                    parent = self.closest_vmec_run(**kwargs)
                    log = logging.getLogger()
                    log.info(
                        "Parent '{parent.vmec_id}' is "
                        "adopting me ('{self.vmec_id}')".format(**locals())
                    )
                    self.parent = parent
                self.run_kwargs[arg] = getattr(self.parent, arg)

    def _start_hpc(self):
        """Start a vmec run via the slurm framework."""

        _max_jobs = 10
        _seconds_to_wait = 60

        log = logging.getLogger()
        folder = os.path.expanduser(self._folder + self.vmec_id + "/")
        input_file = folder + "input." + self.vmec_id
        slurm_file = folder + "slurm_vmec"
        threed1_file = folder + "threed1." + self.vmec_id

        if not os.path.isfile(input_file):
            raise RuntimeError("No input file found")

        if not os.path.isfile(slurm_file):
            raise RuntimeError("No slurm file found")

        queue_cmd = "squeue -u {u} | wc -l".format(u="amerlo")
        cd_cmd = "cd {f}".format(f=folder)
        sbatch_cmd = "sbatch slurm_vmec"
        full_cmd = cd_cmd + " && " + sbatch_cmd

        jobs_in_queue = int(os.popen(queue_cmd).read())
        if not os.path.isfile(threed1_file):
            while jobs_in_queue > _max_jobs:
                log.info(
                    "{j} in the queue, waiting {s} seconds".format(
                        j=jobs_in_queue, s=_seconds_to_wait
                    )
                )
                import time

                time.sleep(_seconds_to_wait)
                jobs_in_queue = int(os.popen(queue_cmd).read())

            log.info("Starting run {r} on HPC backend".format(r=self.vmec_id))
            os.system(full_cmd)

    def start(self, **kwargs):
        """Start a vmec run.

        It will first try to complete the missing run arguments by looking
        at the previous runs stored in the server. The closest run in terms
        of current and pressure profile will then adopt the current run.
        """

        log = logging.getLogger()
        log.info("Preparing run {0} for start.".format(self.vmec_id))

        if self._backend == "hpc":
            return self._start_hpc()

        self._complete_run_kwargs(**kwargs)
        input_data = vmec_input(**self.run_kwargs)
        vmec_server = w7x.get_server(w7x.Server.addr_vmec_server)
        log.info("Starting new VMEC run ({0}).".format(self.vmec_id))
        log.info(str(self))
        vmec_server.service.execVmec(input_data, self.vmec_id)

    def shift_axis(self, shift=-0.01):
        """
        Translate the magneticAxis radially by <shift>
        """
        log = logging.getLogger()
        self.magneticAxis.RCos.coefficients[0] += shift
        log.info("Shifted axis by {shift}.".format(**locals()))

    def converge(self, iteration=0, reducedTolerance=False):
        """
        Returns:
            Run instance of converged run
        """
        log = logging.getLogger()
        if not self.exists():
            self.start()
        else:
            log.info("VMEC run ({0}) already done.".format(self.vmec_id))
        self.wait()
        if not self.was_successful():
            vmec_server = w7x.get_server(w7x.Server.addr_vmec_server)
            threed1Content = vmec_server.service.getVmecRunData(self.vmec_id, "threed1")
            log.info(
                "VMEC run (self.vmec_id) was not successful. - iteration"
                ": {iteration}".format(**locals())
            )
            run_kwargs = copy.deepcopy(self.run_kwargs)
            changed = False
            if "Plasma Boundary exceeded Vacuum Grid Size" in threed1Content:
                max_toroidal_magnetic_flux = (
                    run_kwargs.get(
                        "maxToroidalMagneticFlux",
                        w7x.Defaults.VMEC.maxToroidalMagneticFlux,
                    )
                    * 2.0
                    / 3
                )
                run_kwargs["maxToroidalMagneticFlux"] = max_toroidal_magnetic_flux
                log.info(
                    "Plasma Boundary exceeded Vacuum Grid Size. "
                    "New phiEdge: {max_toroidal_magnetic_flux}".format(**locals())
                )
                changed = True
            if iteration < 2 and "Try increasing NITER" in threed1Content:
                run_kwargs["maxIterationsPerSequence"] = (
                    run_kwargs.get(
                        "maxIterationsPerSequence",
                        w7x.Defaults.VMEC.maxIterationsPerSequence,
                    )
                    + 50000
                )  # noqa
                log.info(
                    "Try increasing NITER. "
                    "New niter: {maxIterationsPerSequence}".format(
                        maxIterationsPerSequence=run_kwargs["maxIterationsPerSequence"]
                    )
                )  # noqa
                changed = True
            if iteration == 2 and reducedTolerance:
                raise ValueError(
                    "Even the tolerance reduced run did not converge."
                )  # noqa
            if iteration >= 2:
                log.info("Still not converged try shrinking the magneticAxis.")
                roughRunKwargs = copy.deepcopy(run_kwargs)
                forceToleranceLevels = roughRunKwargs.pop(
                    "forceToleranceLevels", w7x.Defaults.VMEC.forceToleranceLevels
                )  # noqa
                lastTolerance = forceToleranceLevels[-1] / 1e2
                if (
                    len(forceToleranceLevels) > 1
                    and lastTolerance < forceToleranceLevels[-2]
                ):  # noqa
                    lastTolerance = forceToleranceLevels[-2]
                forceToleranceLevels[-1] = lastTolerance
                roughRunKwargs["forceToleranceLevels"] = forceToleranceLevels
                log.info("Try with lower tolerance {lastTolerance}".format(**locals()))
                roughRun = self.__class__(**roughRunKwargs)
                if self.parent.was_successful():
                    roughRun.parent = self.parent
                roughRun.converge(reducedTolerance=True)
                log.info("Lower tolerance run converged.")
                magneticAxis = roughRun.run_kwargs.get("magneticAxis")
                magneticAxis.RCos.coefficients[0] -= 0.01
                run_kwargs["magneticAxis"] = magneticAxis
                log.info("Set the magneticAxis 1cm inward.")
                changed = True

            if not changed:
                log.error("No response to non converging run.")
                return self
            newRun = self.__class__(**run_kwargs)
            return newRun.converge(iteration=iteration + 1)
        log.info("Run converged successfully")
        return self

    def exists(self):
        if self.vmec_id is None:
            return False
        if self._backend == "web_service":
            return vmec_id_exists(self.vmec_id)
        elif self._backend == "hpc":
            return os.path.isdir(os.path.expanduser(self._folder))

        return False

    def is_ready(self):
        vmec_server = w7x.get_server(w7x.Server.addr_vmec_server)
        return vmec_server.service.isReady(self.vmec_id)

    def was_successful(self):
        return was_successful(self.vmec_id)

    def wait(self, sleepTime=60):
        while not self.is_ready():
            log = logging.getLogger()
            log.debug("Waiting for VMEC run " "{self.vmec_id}.".format(**locals()))
            time.sleep(sleepTime)

    def getBAxis(self):
        return getBAxis(self.vmec_id)

    def volume_lcfs(self):
        self._parse_threed1()
        return self._threed1["PlasmaVolume"]

    def getConfigAndPressureProfileEstimate(self, b_axis=2.5, tolerance=0.001):
        """
        Args:
            b_axis (float): magneticField strength on axis [T]
        """
        b_axisOld = self.getBAxis()
        f = b_axis / b_axisOld
        if abs(1 - f) < tolerance:
            log = logging.getLogger()
            log.warning(
                "Tolerance of {tolerance} was met " "already.".format(**locals())
            )
        magnetic_config = self.magnetic_config.copy()
        pressure_profile = self.pressure_profile.copy()
        magnetic_config.scale_currents(f)
        pressure_profile.scale(f ** 2)
        return magnetic_config, pressure_profile

    def _parse_input(self, **kwargs):
        """Parse vmec input file.

        Args:
            cached (bool): use cached data for the runs.
        """

        local = self._backend == "hpc"
        cached = kwargs.pop("cached", True)
        if "input" in self._parsed:
            return

        path = rna.path.resolve(
            self._folder + "{self.vmec_id}.input.json".format(**locals())
        )
        input_path = rna.path.resolve(
            self._folder + "{self.vmec_id}/input.{self.vmec_id}".format(**locals())
        )

        #  Look for json file first
        if os.path.exists(path) and cached:
            with open(path) as f:
                self._input = json.load(f)
        #  Look for local input.{vmec_id} one
        elif os.path.exists(input_path) and cached and local:
            with open(input_path, "r") as f:
                iterable = f.read().split("\n")
                content = input_transcoding.read(iterable)
                self._input = content
        #  Look up on the server
        else:
            vmec_server = w7x.get_server(w7x.Server.addr_vmec_server)
            content = vmec_server.service.getVmecRunData(self.vmec_id, "input")
            iterable = content.split("\n")
            content = input_transcoding.read(iterable)
            if cached:
                with open(path, "w") as f:
                    json.dump(content, f)
            self._input = content

        self._parsed.append("input")

    def _gen_folder(self):
        """Generate vmec run folder."""

        folder = rna.path.resolve(self._folder + self.vmec_id)
        if not os.path.isdir(folder):
            os.makedirs(folder)

    #  TODO(@amerlo): Add configuration to config.py
    def _gen_slurm(self, **kwargs):
        """Gen HPC slurm configuration.

        Args:
            nodes (int): nodes slurm config.
            tasks (int): tasks-per-node slurm config.
            userid (String): userid to use to generate srun.

        Returns:
            (Dict): dictionary for HPC slurm configuration.
        """

        slurm_config = {}
        slurm_config["job_name"] = "nn-vmec"
        slurm_config["nodes"] = kwargs.pop("nodes", 2)
        slurm_config["tasks"] = kwargs.pop("tasks", 32)
        slurm_config["mem"] = kwargs.pop("mem", "4G")
        slurm_config["partition"] = kwargs.pop("partition", "express")
        slurm_config["mail_user"] = kwargs.pop("userid", "amerlo")
        slurm_config["mail_type"] = "none"
        slurm_config["vmec_exe"] = w7x.config.Defaults.Paths.vmec
        slurm_config["file"] = self._folder + "/slurm_vmec"
        slurm_config["vmec_input"] = "input.{vmec_id}".format(vmec_id=self.vmec_id)

        return slurm_config

    #  TODO(@amerlo): Add configuration to config.py
    def _gen_mgrid(self):
        """Make mgrid file and link into the folder."""

        folder = os.path.expanduser(self._folder + self.vmec_id)
        if not os.path.isdir(folder):
            raise RuntimeError("{f} does not exists".format(f=folder))

        src = w7x.config.Defaults.Paths.mgrid
        dst = folder + "/mgrid_w7x_nv36_hires.nc"
        if not os.path.islink(dst):
            os.symlink(src, dst)

    def _gen_hpc(self, **kwargs):
        """Generate VMEC run files and folders for HPC backend.

        Args:
            nodes (int): nodes slurm config.
            tasks (int): tasks-per-node slurm config.
            userid (String): userid to use to generate srun.
        """

        self._gen_folder()
        self._gen_input()
        self._gen_mgrid()
        self._dump_slurm(**self._gen_slurm(**kwargs))

    def gen(self, **kwargs):
        """Generate VMEC run files and folders.

        Args:
            (Dict): dictionary of configuration parameters for the backend.
        """

        if self._backend != "hpc":
            raise ValueError("{b} is not supported".format(b=self._backend))

        self._gen_hpc(**kwargs)

    def _build_input(self):
        """Build input dict for vmec run."""

        input_data = w7x.vmec.vmec_input(**self.run_kwargs)

        from collections import OrderedDict

        input_dict = OrderedDict()
        input_dict["LFREEB"] = "T" if input_data.freeBoundary else "F"
        input_dict["MGRID_FILE"] = "'{mgrid}'".format(mgrid=input_data.mgridFile)
        input_dict["EXTCUR"] = input_data.coilCurrents
        input_dict["NVACSKIP"] = 6
        input_dict["DELT"] = input_data.timeStep
        input_dict["TCON0"] = input_data.tcon0
        input_dict["NITER"] = input_data.maxIterationsPerSequence
        input_dict["NSTEP"] = input_data.intervalConvergenceOutput
        input_dict["FTOL_ARRAY"] = input_data.forceToleranceLevels
        input_dict["PHIEDGE"] = input_data.maxToroidalMagneticFlux
        input_dict["NFP"] = input_data.numFieldPeriods
        input_dict["MPOL"] = input_data.numModesPoloidal
        input_dict["NTOR"] = input_data.numModesToroidal
        input_dict["NZETA"] = 36  # input_data.numGridPointsToroidal
        input_dict["NTHETA"] = input_data.numGridPointsPoloidal
        input_dict["NS_ARRAY"] = input_data.numGridPointsRadial
        input_dict["NCURR"] = 0 if input_data.iotaProfile else 1
        input_dict["GAMMA"] = input_data.gamma

        radial = np.linspace(0, 1, input_data.numGridPointsRadial[-1])
        input_dict["PMASS_TYPE"] = "'{pmass_type}'".format(
            pmass_type=input_data.pressureProfile.ProfileType
        )
        if input_data.pressureProfile.ProfileType == "cubic_spline":
            profile = create_profile(
                ProfileType="cubic_spline",
                coefficients=input_data.pressureProfile.coefficients,
            )
            input_dict["AM_AUX_F"] = " ".join([str(c) for c in profile(radial)])
            input_dict["AM_AUX_S"] = " ".join([str(s) for s in radial])
        else:
            input_dict["AM"] = input_data.pressureProfile.coefficients

        input_dict["PCURR_TYPE"] = "'{pcur_type}'".format(
            pcur_type=input_data.toroidalCurrentProfile.ProfileType
        )
        if input_data.toroidalCurrentProfile.ProfileType == "cubic_spline":
            profile = create_profile(
                ProfileType="cubic_spline",
                coefficients=input_data.toroidalCurrentProfile.coefficients,
            )
            input_dict["AC_AUX_F"] = " ".join([str(c) for c in profile(radial)])
            input_dict["AC_AUX_S"] = " ".join([str(c) for c in radial])
        else:
            input_dict["AC"] = input_data.toroidalCurrentProfile.coefficients

        input_dict["PRES_SCALE"] = input_data.pressureScale
        input_dict["CURTOR"] = input_data.totalToroidalCurrent
        input_dict["RAXIS_CC"] = input_data.magneticAxis.RCos.coefficients
        input_dict["ZAXIS_CS"] = input_data.magneticAxis.ZSin.coefficients

        r_cos = input_data.boundary.RCos
        z_sin = input_data.boundary.ZSin

        i = 0

        for m in r_cos.poloidalModeNumbers:
            for n in r_cos.toroidalModeNumbers:
                input_dict["RBC({n},{m})".format(**locals())] = r_cos.coefficients[i]
                input_dict["ZBS({n},{m})".format(**locals())] = z_sin.coefficients[i]
                i += 1

        return input_dict

    def _gen_input(self):
        """Generate input.{vmec_id} file."""

        path = rna.path.resolve(
            self._folder + self.vmec_id + "/input.{self.vmec_id}".format(**locals())
        )

        run_content = self._build_input()
        with open(path, "w") as f:
            f.write("&INDATA ")
            for key, value in run_content.items():
                value = (
                    " ".join([str(x) for x in value])
                    if isinstance(value, list)
                    else str(value)
                )
                f.write(key + " = " + value + "\n")
            f.write("/")

    def _dump_slurm(self, **kwargs):
        """Generate slurm_vmec batch file for Draco/Cobra HPC."""

        path = os.path.expanduser(self._folder + self.vmec_id + "/slurm_vmec")

        from collections import OrderedDict

        content = OrderedDict()
        content["shebang"] = "#!/bin/bash -l"
        content["output"] = "#SBATCH -o ./log.txt"
        content["error"] = "#SBATCH -e ./log.txt"
        content["wd"] = "#SBATCH -D ./"
        content["name"] = "#SBATCH -J {job_name}".format(**kwargs)
        content["queue"] = "#SBATCH --partition={partition}".format(**kwargs)
        content["nodes"] = "#SBATCH --nodes={nodes}".format(**kwargs)
        content["tasks"] = "#SBATCH --tasks-per-node={tasks}".format(**kwargs)
        content["mem"] = "#SBATCH --mem={mem}".format(**kwargs)
        content["mail_type"] = "#SBATCH --mail-type={mail_type}".format(**kwargs)
        content["mail_user"] = "#SBATCH --mail-user={mail_user}".format(**kwargs)
        content["time"] = "#SBATCH --time=0:05:00"
        content["srun"] = "srun {vmec_exe} {vmec_input}".format(**kwargs)

        with open(path, "w") as f:
            for line in content.values():
                f.write(line + "\n")

    def _parse_wout(self, **kwargs):
        """Retrieve wout file and populate _wout object attribute.

        Args:
            cached (bool): use cached data for the runs.
        """
        import netCDF4

        if "wout" in self._parsed:
            return

        local = self._backend == "hpc"

        #  TODO(@dobe): Implement wout parsing from web service
        if not local:
            if not self.is_ready():
                raise ValueError(
                    "Calculation with vmec_id {self.vmec_id} has not finishe yet.".format(
                        **locals()
                    )
                )
            return {}

        wout_path = rna.path.resolve(
            self._folder + self.vmec_id + "/wout_{self.vmec_id}.nc".format(**locals())
        )
        if not os.path.exists(wout_path):
            raise ValueError(
                "Calculation with vmec_id {self.vmec_id} has not been "
                "completed".format(**locals())
            )

        #  TODO(@amerlo): Add webservices support
        #  Look for wout_{vmec_id}.nc file
        ds = netCDF4.Dataset(wout_path)
        wout = ds.variables
        self._wout = wout

        self._parsed.append("wout")

    def _parse_threed1(self, **kwargs):
        """Parse threed1 vmec output file.

        Args:
            cached (bool): use cached data for the runs.
        """

        local = self._backend == "hpc"
        cached = kwargs.pop("cached", True)

        if not local:
            if not self.is_ready():
                raise ValueError(
                    "Calculation with vmec_id {self.vmec_id} has not finishe yet.".format(
                        **locals()
                    )
                )

        if "threed1" in self._parsed:
            return

        filePath = rna.path.resolve(
            self._folder + "{self.vmec_id}.threed1.json".format(**locals())
        )
        threed1_path = rna.path.resolve(
            self._folder + self.vmec_id + "/threed1.{self.vmec_id}".format(**locals())
        )

        if cached and local:
            if not os.path.exists(threed1_path):
                raise ValueError(
                    "Calculation with vmec_id {self.vmec_id} has not been "
                    "completed".format(**locals())
                )

        #  Look for the json file first
        if os.path.exists(filePath) and cached:
            with open(filePath) as f:
                self._threed1 = json.load(f)
        #  Look for the threed1.{vmec_id} one
        elif os.path.exists(threed1_path) and cached and local:
            with open(threed1_path, "r") as f:
                iterable = f.read().split("\n")
                content = threed1_transcoding.read(iterable)
                self._threed1 = content
        #  Look up on the server
        else:
            vmec_server = w7x.get_server(w7x.Server.addr_vmec_server)
            content = vmec_server.service.getVmecRunData(self.vmec_id, "threed1")
            iterable = content.split("\n")
            content = threed1_transcoding.read(iterable)
            if cached:
                with open(filePath, "w") as f:
                    json.dump(content, f)
            self._threed1 = content

        self._parsed.append("threed1")

    def to_dict(self, features):
        """Make dict for run.

        Args:
            features (List): list of features to include.

        Returns:
            Dict: dictionary for run.
        """

        if features is None:
            raise ValueError("Features list is empty")

        log = logging.getLogger()

        d = dict.fromkeys(features)
        _features_in_input = [
            "pressure",
            "coilCurrents",
            "phiedge",
            "curtor",
            "toroidalCurrent",
            "ntheta",
            "mpol",
            "ns",
            "pressureScale",
            "pressureProfileCoefficients",
            "toroidalCurrentProfileCoefficients",
        ]
        _features_in_wout = [
            "raxis",
            "zaxis",
            "BCos",
            "RCos",
            "ZSin",
            "LambdaSin",
            "gCos",
            "BsubUCos",
            "BsubVCos",
            "BsubSSin",
            "BsupUCos",
            "BsupVCos",
            "CurrUCos",
            "CurrVCos",
        ]
        _features_in_threed1 = ["totalTime", "PlasmaVolume", "MinorRadius"]

        self._parse_input()
        if not set(features).isdisjoint(set(_features_in_wout)):
            try:
                self._parse_wout()
            except ValueError as e:
                log.warning(e)

        if not set(features).isdisjoint(set(_features_in_threed1)):
            try:
                self._parse_threed1()
            except ValueError as e:
                log.warning(e)

        radial_grid = make_radial_grid(self._input["numGridPointsRadial"])
        dtype = "float32"
        num_coil_currents = 7

        #  TODO(@amerlo): Define get property functions instead
        for f in features:

            #  Check if feature is available
            if f in _features_in_wout and "wout" not in self._parsed:
                value = None
            elif f in _features_in_threed1 and f not in self._threed1.keys():
                value = None
            elif f == "pressure":
                value = self.pressure_profile(radial_grid)
            elif f == "pressureProfileCoefficients":
                value = np.array(self.pressure_profile.coefficients).astype(dtype)
            elif f == "coilCurrents":
                value = np.array(
                    self._input["coilCurrents"][:num_coil_currents]
                ).astype(dtype)
            elif f == "phiedge":
                value = self._input["maxToroidalMagneticFlux"]
            elif f == "curtor":
                value = self._input["totalToroidalCurrent"]
            elif f == "pressureScale":
                value = self._input["pressureScale"]
            elif f == "toroidalCurrent":
                value = self.current_profile(radial_grid)
            elif f == "toroidalCurrentProfileCoefficients":
                value = np.array(self.current_profile.coefficients).astype(dtype)
            elif f == "ntheta":
                value = self._input["ntheta"]
            elif f == "mpol":
                value = self._input["mpol"]
            elif f == "ns":
                value = int(self._input["numGridPointsRadial"][-1])
            elif f == "raxis":
                value = np.array(self._wout["raxis_cc"]).astype(dtype)
            elif f == "zaxis":
                value = np.array(self._wout["zaxis_cs"]).astype(dtype)
            elif f == "BCos":
                value = np.array(self._wout["bmnc"]).astype(dtype)
            elif f == "RCos":
                value = np.array(self._wout["rmnc"]).astype(dtype)
            elif f == "ZSin":
                value = np.array(self._wout["zmns"]).astype(dtype)
            elif f == "LambdaSin":
                value = np.array(self._wout["lmns"]).astype(dtype)
            elif f == "gCos":
                value = np.array(self._wout["gmnc"]).astype(dtype)
            elif f == "BsubUCos":
                value = np.array(self._wout["bsubumnc"]).astype(dtype)
            elif f == "BsubVCos":
                value = np.array(self._wout["bsubvmnc"]).astype(dtype)
            elif f == "BsubSSin":
                value = np.array(self._wout["bsubsmns"]).astype(dtype)
            elif f == "BsupUCos":
                value = np.array(self._wout["bsupumnc"]).astype(dtype)
            elif f == "BsupVCos":
                value = np.array(self._wout["bsupvmnc"]).astype(dtype)
            elif f == "CurrUCos":
                value = np.array(self._wout["currumnc"]).astype(dtype)
            elif f == "CurrVCos":
                value = np.array(self._wout["currvmnc"]).astype(dtype)
            elif f == "totalTime":
                value = self._threed1["totalTime"]
            elif f == "PlasmaVolume":
                value = self._threed1["PlasmaVolume"]
            elif f == "MinorRadius":
                value = self._threed1["MinorRadius"]
            else:
                raise RuntimeError("Feature {f} is not supported".format(**locals()))

            d[f] = value

        return d

    def getInputMagneticAxis(self):
        self._parse_input()
        rac, zas = zip(self._input["value0"], self._input["value1"])
        rcos = FourierCoefficients(
            coefficients=list(rac),
            poloidalModeNumbers=[0],
            toroidalModeNumbers=[0, 1],
            numRadialPoints=1,
        )
        zsin = FourierCoefficients(
            coefficients=list(zas),
            poloidalModeNumbers=[0],
            toroidalModeNumbers=[0, 1],
            numRadialPoints=1,
        )
        magneticAxis = SurfaceCoefficients(RCos=rcos, ZSin=zsin)
        return magneticAxis

    def getMagneticAxis(self):
        self._parse_threed1()
        rcos = FourierCoefficients(
            coefficients=self._threed1["rac"][:2],
            poloidalModeNumbers=[0],
            toroidalModeNumbers=[0, 1],
            numRadialPoints=1,
        )
        zsin = FourierCoefficients(
            coefficients=self._threed1["zas"][:2],
            poloidalModeNumbers=[0],
            toroidalModeNumbers=[0, 1],
            numRadialPoints=1,
        )
        magneticAxis = SurfaceCoefficients(RCos=rcos, ZSin=zsin)
        return magneticAxis

    def getInputBoundary(self):
        self._parse_input()
        coeffDict = {
            (m, n): (rbc, zbs)
            for m, n, rbc, zbs in zip(
                self._input["m"],
                self._input["n"],
                self._input["rbc"],
                self._input["zbs"],
            )
        }

        toroidalModeNumbers = sorted(set([x for x in self._input["n"] if abs(x) <= 6]))
        poloidalModeNumbers = sorted(set([x for x in self._input["m"] if abs(x) <= 6]))
        rbcChoice = []
        zbsChoice = []
        for m in poloidalModeNumbers:
            for n in toroidalModeNumbers:
                if m == 0 and n < 0:
                    rbcChoice.append(coeffDict[(m, abs(n))][0])
                    zbsChoice.append(-coeffDict[(m, abs(n))][1])
                else:
                    rbcChoice.append(coeffDict[(m, n)][0])
                    zbsChoice.append(coeffDict[(m, n)][1])

        rcos = FourierCoefficients(
            coefficients=rbcChoice,
            poloidalModeNumbers=poloidalModeNumbers,
            toroidalModeNumbers=toroidalModeNumbers,
            numRadialPoints=1,
        )
        zsin = FourierCoefficients(
            coefficients=zbsChoice,
            poloidalModeNumbers=poloidalModeNumbers,
            toroidalModeNumbers=toroidalModeNumbers,
            numRadialPoints=1,
        )
        boundary = SurfaceCoefficients(RCos=rcos, ZSin=zsin)
        return boundary

    def beta(self):
        self._parse_threed1()
        return self._threed1["forceIter"][-1]["BETA"][-1]

    def i_tor(self):
        self._parse_input()
        return self._input["i_tor"]

    def force(self):
        self._parse_threed1()
        forceIter = self._threed1["forceIter"][-1]
        attrs = ["FSQR", "FSQZ", "FSQL"]
        return max([forceIter[attr][-1] for attr in attrs])

    def phi_edge_estimate(self, min_volume=25, max_volume=30, island_width=0.075):
        """
        Args:
            min_volume(float): minimal volume in m^3
            max_volume(float): maximal volume in m^3
            island_width(float): radial width of the major islands in m.
                needed to find the maximal radius in order to
                find flux surfaces that are not bound by major islands
        """
        self._parse_threed1()
        p0, p1, p2, p3, p4 = parameters("p0, p1, p2, p3, p4")

        """
        Fit iota(r) with known plasma volume -> V = 2 pi^2 R r^2
        """
        R = self._threed1["MajorRadius"]
        rmaxCalc = np.sqrt(self.volume_lcfs() / (2 * np.pi ** 2 * R))
        rValues = np.array(self._threed1["S"]) * rmaxCalc
        iotaValues = self._threed1["IOTA"]
        fluxValues = self._threed1["TOROIDAL_FLUX"]

        iota = Variable("iota")
        r = Variable("r")
        rPoly = p0 + p1 * r + p2 * r ** 2 + p3 * r ** 3 + p4 * r ** 4
        iotaPoly = p0 + p1 * iota + p2 * iota ** 2 + p3 * iota ** 3 + p4 * iota ** 4
        iotaVsR = fit(rPoly, rValues, iotaValues)
        rVsIota = fit(iotaPoly, iotaValues, rValues)

        """
        Find max r
            -> greater than min_volume
            -> smaller than max_volume
            -> not bound by major islands
        """
        rmaxUser = np.sqrt(max_volume / (2 * np.pi ** 2 * R))
        rminUser = np.sqrt(min_volume / (2 * np.pi ** 2 * R))

        if rmaxCalc < rminUser:
            rmaxCalc = 0.6 * (rmaxUser + rminUser)
        rmax = min((rmaxCalc, rmaxUser))
        iIsland = [4.0 / 5.0, 5.0 / 5.0, 6.0 / 5.0]
        rIsland = [rVsIota.subs(iota, i) for i in iIsland]
        for ri in rIsland:
            if abs(rmax - ri) < 0.5 * island_width:
                rmax = ri - 0.5 * island_width
                break

        """
        fit flux(iota)
        """
        toroidalFlux = p0 + p1 * iota + p2 * iota ** 2 + p3 * iota ** 3 + p4 * iota ** 4
        fluxVsIota = fit(toroidalFlux, iotaValues, fluxValues)

        iotaMax = iotaVsR.subs(r, rmax)
        return fluxVsIota.subs(iota, iotaMax)


def getP0Estimate(beta_scan_runs, beta, normTolerance=500):
    """Make shure, you fit on the same bField strenghts."""
    norms = [r.magnetic_config.coil_currents("Aw")[0] for r in beta_scan_runs]
    if not len(set(norms)) < 2:
        raise ValueError(
            "Runs of beta scan have difference coilCurrent norms"
            "({norms})".format(**locals())
        )

    p0s = [0] + [r.pressure_profile.norm for r in beta_scan_runs]
    betas = [0] + [r.beta() for r in beta_scan_runs]

    a, b = parameters("a, b")
    x, y = variables("x, y")
    model = Model({y: a * x + b})
    fit = Fit(model, x=p0s, y=betas)
    res = fit.execute()
    fun = model[y]
    fun = fun.subs(a, res.params["a"])
    fun = fun.subs(b, res.params["b"])
    iFun = sympy.solve(y - fun, x)[0]
    p0 = iFun.subs(y, beta)

    return p0


def adjust_beta(
    currents_rw,
    beta,
    b_axis=2.5,
    rel_pressure_profile=[1, -1],
    rel_current_profile=[0, 1, -1],
    totalToroidalCurrent=None,
    beta_precision=0.001,
    beta_scan_runs=None,
):
    """
    Args:
        currents_rw (parent or relativeCurrents): 7 ratios (1 winding emulation)
        beta (float): beta = 2 * mu_0 * <p> / <B^2>
        b_axis (float): B_toroidal on the axis (s=0). Usually: 2.5
        dIota (float): delta iota to one of the three edge resonances (4/5, 5/5,
            6/5
    """
    log = logging.getLogger()

    """0) Prepare"""
    if beta_scan_runs is None:
        beta_scan_runs = []
    else:
        beta_scan_runs = [Run(r) for r in beta_scan_runs]
    if len(beta_scan_runs) > 0:
        norm = beta_scan_runs[0].magnetic_config.coil_currents("Aw")[0]
    else:
        norm = 12000 * 108

    magnetic_config = w7x.MagneticConfig.from_currents(
        *currents_rw, unit="rw", scale=norm
    )
    current_profile = PowerSeries(coefficients=rel_current_profile)
    currentKwargs = dict(
        current_profile=current_profile, totalToroidalCurrent=totalToroidalCurrent
    )

    betaScanRunKwargs = dict(
        magnetic_config=magnetic_config,
        maxIterationsPerSequence=60000,
        numGridPointsRadial=[4, 9, 28, 51],
        forceToleranceLevels=[1e-3, 1e-5, 1e-9, 1e-11],
    )
    betaScanRunKwargs.update(currentKwargs)

    if len(beta_scan_runs) == 0:
        """0) Run initial beta scan run"""
        pressure_profile = PowerSeries(
            coefficients=[6e4 * x for x in rel_pressure_profile]
        )
        run = Run(pressure_profile=pressure_profile, **betaScanRunKwargs)
        run.converge()
        if not run.was_successful():
            raise RuntimeError("First run was not successful.")
        else:
            beta_scan_runs.append(run)

    pressure_profile = PowerSeries(coefficients=rel_pressure_profile)
    """1) Find correct p0 """
    log.info("1) estimate p0")
    matching_beta_runs = {}
    for run in beta_scan_runs:
        # any run already has the correct beta?
        if abs(run.beta() - beta) < beta_precision:
            matching_beta_runs[run.force()] = run
    if matching_beta_runs:
        # get most precise run from the matching ones
        grand_parent = matching_beta_runs[min(matching_beta_runs.keys())]
        log.info(
            "Found run with correct beta in given runs: "
            "{grand_parent.vmec_id}".format(**locals())
        )
    else:
        # approach correct beta.
        pDev = [pressure_profile.deviation(r.pressure_profile) for r in beta_scan_runs]
        greatGrandParent = beta_scan_runs[pDev.index(min(pDev))]
        while True:
            p0 = getP0Estimate(beta_scan_runs, beta=beta)
            phiEdge = greatGrandParent.phi_edge_estimate()
            pressure_profile.norm = p0
            grand_parent = Run(
                parent=greatGrandParent,
                pressure_profile=pressure_profile,
                maxToroidalMagneticFlux=phiEdge,
                **betaScanRunKwargs
            )
            log.info(
                "Start run with p0 estimate: {p0}. "
                "Also adjust phiEdge to {phiEdge}".format(**locals())
            )
            grand_parent.converge()
            foundBeta = grand_parent.beta()
            if abs(foundBeta - beta) < beta_precision:
                log.info(
                    "Retrieved beta({foundBeta}) matches "
                    "Requested({beta}) within tolerance.".format(**locals())
                )
                break
            else:
                log.info(
                    "Retrieved beta({foundBeta}) not close enough to "
                    "Requested({beta}).".format(**locals())
                )
                greatGrandParent = grand_parent
                beta_scan_runs.append(greatGrandParent)

    """2) Become more precise """
    log.info("2) calculate more precise")
    force_tolerance = 1e-12
    if grand_parent.force() < force_tolerance:
        parent = grand_parent
        log.info("Forces in grand_parent are sufficiently small. ")
    else:
        parent = grand_parent.mutate(
            maxIterationsPerSequence=60000,
            numGridPointsRadial=[4, 9, 28, 51],
            forceToleranceLevels=[1e-3, 1e-5, 1e-9, force_tolerance],
        )
        parent.converge(iteration=2)

    """3) Find phiEdge and magnetic_config/p0 for correct Volume/b_axis """
    log.info(
        "3) Get phiEdge and magnetic_config/p0 estimate for correct " "Volume/b_axis"
    )
    phiEdge = -1.8  # parent.phi_edge_estimate()
    magnetic_config, pressure_profile = parent.getConfigAndPressureProfileEstimate()

    """4) Run again with parameters found"""
    log.info(
        "4) converge run with phiEdge = {phiEdge} and magnetic_config"
        "to reach b_axis = {b_axis}".format(**locals())
    )
    final_force_tolerance = 1e-14
    final_run = Run(
        maxToroidalMagneticFlux=phiEdge,
        maxIterationsPerSequence=100000,
        magnetic_config=magnetic_config,
        pressure_profile=pressure_profile,
        forceToleranceLevels=[1e-3, 1e-5, 1e-9, final_force_tolerance],
        parent=parent,
        **currentKwargs
    )
    final_run.converge()
    log.info(str(final_run))
    return final_run


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm


def plot_fourier(fourier, **kwargs):
    """Plot fourier coefficients.

    Args:
        fourier (w7x.vmec.FourierCoefficients): fourier coefficients to plot.

    Returns:
        matplotlib.axes: axes.
    """

    num_radial_points = fourier.numRadialPoints
    vmin = kwargs.pop("vmin", 1e-6)
    vmax = kwargs.pop("vmax", 1)
    cax = kwargs.pop("cax", None)
    orientation = kwargs.pop("orientation", "horizontal")

    #  TODO(@amerlo): Index coefficiets based on radial location
    if num_radial_points != 1:
        raise RuntimeError(
            "Plot for {n} is not implemented yet".format(n=num_radial_points)
        )

    ax = kwargs.pop("axes", None)
    shape = (len(fourier.poloidalModeNumbers), len(fourier.toroidalModeNumbers))
    coeffs = np.reshape(fourier.coefficients, shape)

    img = ax.imshow(
        np.absolute(coeffs),
        origin="lower",
        norm=LogNorm(vmin=vmin, vmax=vmax),
        cmap=plt.get_cmap("YlGnBu_r"),
    )

    if cax is not None:
        plt.colorbar(img, cax=cax, orientation=orientation)

    ax.set_yticks(fourier.poloidalModeNumbers)
    ax.set_xticks(np.arange(0, len(fourier.toroidalModeNumbers)))
    ax.set_xticklabels(fourier.toroidalModeNumbers)
    ax.set_yticklabels(fourier.poloidalModeNumbers)

    return ax


def plot_surface(surface, phi, scalar=None, **kwargs):
    """Plot fourier surface.

    Args:
        surface (w7x.vmec.SurfaceCoefficients): surface object to plot.
        phi (float): toroidal angle to plot in rad.
        scalar (w7x.vmec.FourierCoefficients): scalar values to plot on top.

    Returns:
        matplotlib.axes: axes.
    """

    surface_radial_points = surface.RCos.numRadialPoints

    ax = kwargs.pop("axes", None)
    levels = kwargs.pop("levels", 20)
    cmap = kwargs.pop("cmap", "RdBu_r")

    num_radial_points = kwargs.pop("nRad", 10)
    num_poloidal_points = kwargs.pop("mPol", 36)
    poloidal_grid = np.linspace(0, 2 * np.pi, num_poloidal_points)

    if surface_radial_points < num_radial_points:
        num_radial_points = surface_radial_points

    component = kwargs.pop("component", None)

    r = []
    z = []
    values = []

    for s in np.linspace(1, surface_radial_points - 1, num_radial_points, dtype=int):
        r.extend(list(surface(s, phi, poloidal_grid)[0]))
        z.extend(list(surface(s, phi, poloidal_grid)[1]))

        if scalar is not None:
            values.extend(list(scalar(s, phi, poloidal_grid, component)))

    if scalar is not None:
        tricnt = ax.tricontourf(r, z, values, levels=levels, cmap=cmap)
        plt.colorbar(tricnt, ax=ax)

    ax.plot(r, z, **kwargs)

    return ax


def _plot_vmec_coil_currents(data, **kwargs):
    """Generate axis for coil currents plot.

    Args:
        data (pandas.DataFrame): data to plot.
        kwargs: Options to pass to matplotlib plotting method.

    Returns:
        matplotlib.axes.Axes: numpy.ndarray of axis, one for each
        coil current to plot.
    """

    _cols = 3
    _rows = 3

    currents = ["i1", "i2", "i3", "i4", "i5", "iA", "iB"]

    # @amerlo: Because pandas is used only here I do not (yet) want to depend on
    # it. I want to keep w7x slim for deployment. Thus I put the import here.
    import pandas as pd

    df = pd.DataFrame(data.values.tolist(), columns=currents, index=data.index)

    _, axs = plt.subplots(_rows, _cols, constrained_layout=True, **kwargs)
    for ax, current in zip(axs.flat, currents):
        ax.set_title(current)
        df[current].plot(kind="hist", ax=ax)

    return axs


def _plot_vmec_profile(data):
    """Plot vmec profile.

    Args:
        data (pandas.DataFrame): data to plot.

    Returns:
        matplotlib.axes.Axes: matplotlib axes.
    """

    number_radial_points = len(data.iloc[0])
    s = np.linspace(0.0, 1.0, number_radial_points)

    return [plt.plot(s, profile) for _, profile in data.items()]


def plot_vmec(data, feature, **kwargs):
    """Plot vmec runs data.

    Args:
        data (pandas.DataFrame): data to plot.
        feature (String): data column to plot.
        kwargs: Options to pass to matplotlib plotting method.

    Returns:
        matplotlib.axes.Axes: matplotlib axis.
    """

    _profiles = ["pressure", "iota", "toroidalCurrent"]

    if feature == "coilCurrents":
        return _plot_vmec_coil_currents(data[feature], **kwargs)
    elif feature in _profiles:
        return _plot_vmec_profile(data[feature])
    else:
        return data[feature].plot(kind="hist", **kwargs)



"""
Abstract vmec bacth object used to generate input file for vmec code
"""

import w7x
import rna
import dill as pickle
import json
import logging
import numpy as np
import pandas as pd
import os
import itertools
import hashlib
import shutil
from sklearn.gaussian_process import GaussianProcessRegressor

#  TODO(@amerlo): Decouple Distribution, Vmec and Batch classes in
#  multiple modules.


#  TODO(@amerlo): Move this to util functions.
def repeatfunc(func, *args):
    """Repeat calls to func with specified arguments.

    Args:
        func (Function): function to repeat.
        args (): arguments to pass to the function.

    Examples:
        >>> import random
        >>> import w7x
        >>> _ = w7x.batch.repeatfunc(random.random)
    """

    return itertools.starmap(func, itertools.repeat(args))


def loguniform(low=0, high=1, size=None, base=10):
    """LogUniform distribution wrapper for numpy.random.uniform."""
    return np.power(base, np.random.uniform(low, high, size))


class Distribution(object):
    """Handle distribution as iterator to more efficiently sample from them.

    Examples:
        >>> import w7x
        >>> import random
        >>> d = w7x.batch.Distribution(random.randint, 0, 10)
        >>> assert isinstance(next(d), int)
    """
    def __init__(self, dist, *args):

        self.dist = repeatfunc(dist, *args)

    def __next__(self):
        return next(self.dist)


class GaussianProcess(object):
    """Gaussian process.

    Examples:
        >>> import w7x
        >>> gp = w7x.batch.GaussianProcess([1.0, 0.0])
        >>> sample = gp()
    """

    _max_seed = 100000
    _grid_points = 128
    _profile_points = np.linspace(0, 1, _grid_points).reshape(-1, 1)

    def __init__(self, y, x=None, gradient=None, **kwargs):

        self.regressor = GaussianProcessRegressor(**kwargs)

        if x is None:
            x = np.linspace(0, 1, len(y)).reshape(-1, 1)
        else:
            if len(x) != len(y):
                raise ValueError(
                    "The given profile must be of equal length of x")
            x = np.array(x).reshape(-1, 1)

        self.x = x
        self.y = np.array(y)

        self.gradient = gradient
        self.x_grid = (max(x) - min(x))[0] / GaussianProcess._grid_points

        self.regressor.fit(self.x, self.y)

    def __call__(self):
        """Samples from the posteriori distribution."""

        if self.gradient is None:
            return self.regressor.sample_y(
                GaussianProcess._profile_points,
                1,
                random_state=np.random.randint(
                    GaussianProcess._max_seed)).flatten()

        # TODO(@amerlo): local variable sample assigned but never used
        # sample = self.regressor.sample_y(
        #     GaussianProcess._profile_points,
        #     1,
        #     random_state=np.random.randint(
        #         GaussianProcess._max_seed)).flatten()

        #  Evaluate virtual observations based on gradiant relationships
        virtual_x = [x + self.x_grid for x in self.x.flatten()
                     ] + [x - self.x_grid
                          for x in self.x.flatten()] + list(self.x.flatten())
        virtual_y = [
            y + grad * self.x_grid
            for y, grad in zip(self.y.flatten(), self.gradient)
        ] + [
            y - grad * self.x_grid
            for y, grad in zip(self.y.flatten(), self.gradient)
        ] + list(self.y.flatten())

        #  Apply virtual observations
        self.regressor.fit(
            np.array(virtual_x).reshape(-1, 1), np.array(virtual_y))

        return self.regressor.sample_y(
            GaussianProcess._profile_points,
            1,
            random_state=np.random.randint(
                GaussianProcess._max_seed)).flatten()


def batch_id_exists(folder, batch_id):
    """Check if {batch_id} batch is located at {folder} location.

    This check returns {True} in case of folder exists, but none check
    is performed on its content.

    Args:
        folder (String): path of the folder where to check.
        batch_id (String): batch id to check.

    Returns:
        Bool: {True} if the batch folder exists.
    """

    path = rna.path.resolve(folder + batch_id)
    return os.path.isdir(path)


class Batch(object):
    """Batch object to handle common batch workflows.

    Examples:
        >>> import w7x
        >>> batch = w7x.batch.Batch(folder='~/tmp/VMEC/test')  # TODO: fix
        Traceback (most recent call last):
        ...
        NotImplementedError


    """
    def __init__(self, *args, **kwargs):

        folder = kwargs.pop('folder', None)
        cached = kwargs.pop('cached', True)

        if folder is None:
            raise ValueError("Please specify batch folder")

        if folder[-1] != "/":
            self.folder = folder + "/"
        else:
            self.folder = folder

        if len(args) != 0:
            batch_id = args[0]
            if batch_id_exists(self.folder, batch_id) and cached:
                self._load_batch(batch_id)
        else:
            self.cached = cached
            self._size = w7x.config.Batch.DatasetConfig.size
            self._iteration = 0
            self._run_counter = 0
            self._runs = []
            self._init_default()
            self.batch_id = self._build_batch_id()
            self.features = None
            self.backend = w7x.config.Hpc.Draco.small

    @property
    def size(self):
        """Return the size of the batch."""
        return self._size

    @size.setter
    def size(self, size):
        """Set the size of the batch."""
        if size <= 0:
            raise ValueError("Size must be a non negative integer")

        if not isinstance(size, int):
            raise ValueError("Size must be an integer value")

        self._size = size

    @property
    def code(self):
        """Return the code name."""
        raise NotImplementedError("Code string should be implemented")

    def _load_batch(self, batch_id):
        """Load batch from pkl file and populate its attribute.

        Args:
            batch_id (String): batch id to load back.
        """

        #  Keep track of original folder
        folder = self.folder

        with open(folder + batch_id + '.pkl', 'rb') as f:
            obj = pickle.load(f)

        for key, value in obj.__dict__.items():
            setattr(self, key, value)

        self.folder = folder

    def _init_default(self):
        """Init default batch."""
        raise NotImplementedError()

    def _build_batch_id(self):
        """Build batch id.

        Returns:
            String: hashed batch id.
        """

        #  TODO(@amerlo): Handle not serializable attributes
        #  Build batch id based on hashed view of the object
        batch = json.dumps(self.__dict__,
                           sort_keys=True,
                           default=lambda o: "type={t}".format(t=type(o)))
        batch_id = ("w7x-{code}-{hash_id}".format(
            code=self.code, hash_id=hashlib.md5(batch.encode()).hexdigest()))

        if batch_id_exists(self.folder, batch_id):
            self._iteration += 1
            return self._build_batch_id()

        return batch_id

    def _to_pickle(self):
        """Dump object to pickle file."""

        path = os.path.expanduser(self.folder + self.batch_id)

        with open(path + '.pkl', 'wb') as f:
            pickle.dump(self, f)

    def _gen(self):
        """Generate one run.

        Returns:
            Object: generated run.
            String: id for the generated run.
        """
        raise NotImplementedError()

    def gen(self):
        """Gen all folders and files required for batch runs."""

        if batch_id_exists(self.folder, self.batch_id) and self.cached:
            raise RuntimeError(
                "Batch {id} already exists".format(id=self.batch_id))

        batch_folder = os.path.expanduser(self.folder + self.batch_id)
        if os.path.isdir(batch_folder):
            shutil.rmtree(batch_folder)
        os.makedirs(batch_folder)

        for i in range(self.size):
            (_, run_id) = self._gen()
            self._runs.append(run_id)

        self._to_pickle()

    def _build_run_id(self):
        """Build hashed run id for current batch.

        Returns:
            String: hashed id for run.
        """

        #  Couple batch and run id
        run_string = self.batch_id + str(self._run_counter)
        run_id = ("w7x-{code}-{hash_id}".format(
            code=self.code,
            hash_id=hashlib.md5(run_string.encode()).hexdigest()))
        self._run_counter += 1

        return run_id

    @property
    def runs(self):
        """Return a list of runs Ids.

        Returns:
            List: list of runs Ids.
        """
        return self._runs

    def _run_to_dict(run_id, feautures):
        """Make dict from run.

        Args:
            run_id (String): run to look for.
            features (List): list of features to retrieve.

        Returns:
            Dict: dictionary for the run.
        """
        raise NotImplementedError()

    def to_dataframe(self):
        """Generate pandas dataframe from batch.

        Returns:
            pandas.DataFrame: pandas dataframe of the batch.
        """

        if self.features is None:
            raise ValueError("Please set at least one feature")

        d = dict.fromkeys(self.runs)
        for run_id in d.keys():
            d[run_id] = self._run_to_dict(run_id, self.features)

        df = pd.DataFrame(d).T

        return df

    def start(self):
        """Start batch execution."""

        batch_folder = os.path.expanduser(self.folder + self.batch_id)

        if not os.path.isdir(batch_folder):
            raise RuntimeError("Batch folder not found")

        for run_id in self.runs:
            self._start(run_id)

    def check(self):
        """Check batch exectuion."""

        batch_folder = os.path.expanduser(self.folder + self.batch_id)

        if not os.path.isdir(batch_folder):
            raise RuntimeError("Batch folder not found")

        logger = logging.getLogger()
        logger.info("%s: checking batch ...", self.batch_id)

        fails = 0
        for run_id in self.runs:
            if self._check(run_id) is not None:
                fails += 1

        logger.info("%s: %d runs found", self.batch_id, len(self.runs))
        logger.info("%s: %d runs failed", self.batch_id, fails)

    @classmethod
    def concat(cls, *args, folder, features):
        """Concat batches into single dataframe.

        Args:
            *args (list): list of batch ids.
            folder (str): folder where batches are located.
            features (list): list of features to retrieve.

        Returns:
            pd.DataFrame: concatenated dataframe.
        """

        folder = os.path.dirname(folder)
        runs = pd.DataFrame()

        for batch_id in list(*args):

            #  TODO(@amerlo): How to select which code they are?
            batch = cls(batch_id, folder=folder)
            batch.features = features
            batch_df = batch.to_dataframe()

            #  Append batch_id to all batch runs
            batch_df["batch"] = [batch_id] * len(batch_df)

            #  Check for run ids overlapping
            if not set(batch_df.index).isdisjoint(set(runs.index)):
                raise RuntimeError("Duplicated run ids found")

            runs = pd.concat([runs, batch_df])

        return runs


class Vmec(Batch):
    """Create a batch of VMEC runs.

    This class does support only the HPC cluster backend and it will
    generate input file compatible with this backend.

    Examples:
        >>> import w7x
        >>> vmec_batch = w7x.batch.Vmec(folder='~/tmp/VMEC/test')
        >>> vmec_batch.size
        1000
    """

    #  TODO(@amerlo): Armonize where to store default parameters
    def __init__(self, *args, **kwargs):
        super(Vmec, self).__init__(*args, **kwargs)

    def _init_default(self):
        """Init default batch for VMEC."""

        self.magnetic_config = (w7x.Defaults.MagneticConfig.standard_rw, 'rw')
        self.pressure_profile = ([1e-6, -1e-6], 'power_series')
        self.current_profile = ([1e-6, -1e-6], 'power_series')

        self.phiedge = w7x.config.Defaults.VMEC.maxToroidalMagneticFlux
        self.curtor = 0.0
        self.pressure_scale = 1.0

        #  Fixed runtime parameters
        self.magnetic_axis = AxisConfig.nn_axis
        self.boundary = BoundaryConfig.nn_boundary
        self.niter = w7x.config.Defaults.VMEC.maxIterationsPerSequence
        self.ntheta = w7x.config.Defaults.VMEC.ntheta
        self.nmodes = w7x.config.Defaults.VMEC.numModesPoloidal
        self.nradial = w7x.config.Defaults.VMEC.numGridPointsRadial
        self.tlevels = w7x.config.Defaults.VMEC.forceToleranceLevels
        self.initial_guess = False

    @property
    def code(self):
        """Return code name, 'vmec' in this case."""
        return 'vmec'

    @property
    def magnetic_config(self):
        return w7x.MagneticConfig.from_currents(*next(self._coil_currents),
                                                unit=self._coil_currents_unit)

    @magnetic_config.setter
    def magnetic_config(self, magnetic_config):
        """Set magnetic configuration.

        Args:
            magnetic_config (List, string): list of coil currents, currents unit.
        """

        currents, unit = magnetic_config

        if currents.__class__.__name__ != 'Distribution':
            currents = Distribution(lambda x: x, currents)
        self._coil_currents = currents

        self._coil_currents_unit = unit

    @property
    def pressure_profile(self):
        """Return normalized pressure profile.

        Returns:
            w7x.vmec.Profile: normalized profile.
        """

        return w7x.vmec.create_profile(
            ProfileType=self._pressure_profile_type,
            coefficients=next(self._pressure_profile_coefficients))

    @pressure_profile.setter
    def pressure_profile(self, pressure_profile):
        """Set pressure profile.

        Args:
            pressure_profile (List, string): list of coefficients, type.
        """

        coeffs, typ = pressure_profile

        if coeffs.__class__.__name__ != 'Distribution':
            coeffs = Distribution(lambda x: x, coeffs)
        self._pressure_profile_coefficients = coeffs

        self._pressure_profile_type = typ

    @property
    def current_profile(self):
        profile = w7x.vmec.create_profile(
            ProfileType=self._current_profile_type,
            coefficients=next(self._current_profile_coefficients))
        if profile.ProfileType == 'power_series' and self._current_profile_sanitize:
            profile.set_value(1.0, 0.0)
        return profile

    @current_profile.setter
    def current_profile(self, current_profile):
        """Set current profile.

        Args:
            current_profile (List, string, bool):
                list of coefficients, type, sanitize.
        """

        try:
            coeffs, typ, sanitize = current_profile
        except ValueError:
            coeffs, typ = current_profile
            sanitize = True

        if coeffs.__class__.__name__ != 'Distribution':
            coeffs = Distribution(lambda x: x, coeffs)
        self._current_profile_coefficients = coeffs

        self._current_profile_type = typ
        self._current_profile_sanitize = sanitize

    @property
    def nmodes(self):
        return int(next(self._nmodes))

    @nmodes.setter
    def nmodes(self, nmodes):
        """Set number of poloidal and toroidal modes.

        Args:
            nmodes (int): number of fourier modes to use.
        """

        if nmodes.__class__.__name__ != 'Distribution':
            nmodes = Distribution(lambda x: x, nmodes)
        self._nmodes = nmodes

    @property
    def phiedge(self):
        return next(self._phiedge)

    @phiedge.setter
    def phiedge(self, phiedge):
        """Set the maximum toroidal magnetic flux.

        Args:
            phiedge (float): phiedge value.
        """

        if phiedge.__class__.__name__ != 'Distribution':
            phiedge = Distribution(lambda x: x, phiedge)
        self._phiedge = phiedge

    @property
    def curtor(self):
        return next(self._curtor)

    @curtor.setter
    def curtor(self, curtor):
        """Set the total toroidal current.

        Args:
            curtor (float): curtor value.
        """

        if curtor.__class__.__name__ != 'Distribution':
            curtor = Distribution(lambda x: x, curtor)
        self._curtor = curtor

    @property
    def pressure_scale(self):
        return next(self._pressure_scale)

    @pressure_scale.setter
    def pressure_scale(self, pressure_scale):
        """Set the scaling for the pressure profile.

        Args:
            pressure_scale (float): scaling for the pressure profile.
        """

        if pressure_scale.__class__.__name__ != 'Distribution':
            pressure_scale = Distribution(lambda x: x, pressure_scale)
        self._pressure_scale = pressure_scale

    @property
    def ntheta(self):
        return int(next(self._ntheta))

    @ntheta.setter
    def ntheta(self, ntheta):
        """Set ntheta.

        Args:
            ntheta (int): ntheta value.
        """

        if ntheta.__class__.__name__ != 'Distribution':
            ntheta = Distribution(lambda x: x, ntheta)
        self._ntheta = ntheta

    def _sample_run_parameters(self):
        """Sample run parameters from object.

        Returns:
            dict: dict of input runs parameters.
        """

        input_data = {}

        input_data['magnetic_config'] = self.magnetic_config
        input_data['pressure_profile'] = self.pressure_profile
        input_data['current_profile'] = self.current_profile

        input_data['numModesPoloidal'] = self.nmodes
        input_data['numModesToroidal'] = input_data['numModesPoloidal']
        input_data['magneticAxis'] = self.magnetic_axis
        input_data['boundary'] = self.boundary
        input_data['maxToroidalMagneticFlux'] = self.phiedge
        input_data['pressureScale'] = self.pressure_scale
        input_data['totalToroidalCurrent'] = self.curtor
        input_data['numGridPointsRadial'] = self.nradial
        input_data['forceToleranceLevels'] = self.tlevels
        input_data['numGridPointsPoloidal'] = self.ntheta
        input_data['maxIterationsPerSequence'] = self.niter
        input_data['folder'] = self.folder + self.batch_id + "/"
        input_data['getInitialGuess'] = self.initial_guess

        return input_data

    #  @TODO(amerlo): Check if feasible to have full run object in memory
    def _gen(self):
        """Generate VMEC run.

        Returns:
            Object: generated VMEC run.
            String: VMEC run id.
        """

        backend = self.backend['name']
        run_input_parameters = self._sample_run_parameters()
        gen_input_parameters = self.backend

        run_id = self._build_run_id()
        run = w7x.vmec.Run(run_id, backend=backend, **run_input_parameters)
        run.gen(**gen_input_parameters)

        return (run, run_id)

    def _run_to_dict(self, run_id, features):
        """Make dict from run.

        Args:
            run_id (String): run to look for.
            features (List): list of features to retrieve.

        Returns:
            Dict: dictionary for the run.
        """

        folder = self.folder + self.batch_id + "/"
        run = w7x.vmec.Run(run_id, folder=folder, backend='hpc')

        return run.to_dict(features)

    def _start(self, run_id):
        """Start vmec run.

        Args:
            run_id (String): vmec run id to start.
        """

        folder = self.folder + self.batch_id + "/"
        run = w7x.vmec.Run(run_id, backend='hpc', folder=folder)
        run.start()

    def _check(self, run_id):
        """Check vmec run.

        Args:
            run_id (str): vmec run id to check.
        """

        logger = logging.getLogger()
        logger.info("%s: checking run ...", run_id)

        folder = self.folder + self.batch_id + "/" + run_id

        input_file = folder + "/input." + run_id
        threed1_file = folder + "/threed1." + run_id
        wout_file = folder + "/wout_" + run_id + ".nc"

        if not os.path.isfile(input_file):
            logger.info("%s: input file not found", run_id)
            return

        if not os.path.isfile(threed1_file):
            logger.info("%s: threed1 file not found", run_id)
            return

        if os.path.isfile(wout_file):
            return

        #  TODO(@amerlo): Implement error proxies
        #  TODO(@amerlo): Implement more error messages
        logger.info("%s: wout file not found", run_id)
        with open(threed1_file) as f:
            if 'Plasma Boundary exceeded Vacuum Grid Size' in f.read():
                logger.info("%s: plasma boundary exceeded", run_id)
                return -3
            #  Fall back to time limit if no error messages have been found
            if 'EXECUTION TERMINATED NORMALLY' not in f.read():
                logger.info("%s: time limit reached", run_id)
                return -2

        #  "-1" is a generic error
        logger.info(
            "%s: execution terminated normally but no "
            "wout file has been found")
        return -1

    def plot(self, *args, **kwargs):
        """Plot distribution of requested batch parameters.

        For not supported features, it will tries to use pandas
        DataFrame plot() function.

        Args:
            *args: feature to plot.
            num (int): number of runs to plot.
            **kwargs: keywords arguments to pass to matplotlib.

        Returns:
            matplotlib.axes.Axes: If the backend is not the default matplotlib one,
            the return value will be the object returned by the backend.
        """

        if len(args) == 0:
            raise ValueError("Please provide one feature to plot")

        if len(args) > 1:
            raise ValueError("Only one feature plot is supported")

        df = self.to_dataframe()
        df.dropna(inplace=True)
        df.drop(df.columns.difference(args), axis=1, inplace=True)

        num = kwargs.pop('num', None)
        if num is not None:
            drop_indices = np.random.choice(df.index,
                                            len(df) - num,
                                            replace=False)
            df.drop(drop_indices, axis=0, inplace=True)

        return w7x.plotting.vmec.plot_vmec(df, *args, **kwargs)


#  TODO(@amerlo): find good default value for different batches
class AxisConfig:
    r_cos = [5.52, 0.28143]
    z_sin = [0.0, -0.23796]

    nn_axis = w7x.vmec.SurfaceCoefficients(
        RCos=w7x.vmec.FourierCoefficients(coefficients=r_cos,
                                          poloidalModeNumbers=[0],
                                          toroidalModeNumbers=[0, 1],
                                          numRadialPoints=1),
        ZSin=w7x.vmec.FourierCoefficients(coefficients=z_sin,
                                          poloidalModeNumbers=[0],
                                          toroidalModeNumbers=[0, 1],
                                          numRadialPoints=1))


class BoundaryConfig:
    r_cos = [0.0, 5.52, 0.28143, 0.029677, 0.48552, -0.23402]
    z_sin = [0.0, 0.0, -0.23796, 0.032068, 0.62135, 0.1838]

    nn_boundary = w7x.vmec.SurfaceCoefficients(
        RCos=w7x.vmec.FourierCoefficients(coefficients=r_cos,
                                          poloidalModeNumbers=[0, 1],
                                          toroidalModeNumbers=[-1, 0, 1],
                                          numRadialPoints=1),
        ZSin=w7x.vmec.FourierCoefficients(coefficients=z_sin,
                                          poloidalModeNumbers=[0, 1],
                                          toroidalModeNumbers=[-1, 0, 1],
                                          numRadialPoints=1))


if __name__ == '__main__':
    import doctest
    doctest.testmod()



from six import string_types


class GeoSet(list):
    def __init__(self, mm_ids=[], name="mm_id", color="black", paths=None):
        self.name = name
        self.color = color
        self.paths = paths
        self.hits = 0
        self.load = 0.0
        content = []
        for mm_id in mm_ids:
            if isinstance(mm_id, int):
                content.append(mm_id)
            elif isinstance(mm_id, GeoSet):
                content.extend(mm_id)
            elif isinstance(mm_id, string_types):
                from w7x import MeshedModelsIds

                content.extend(getattr(MeshedModelsIds, mm_id))
            else:
                raise NotImplementedError("mm_id {mm_id}".format(**locals()))
        super(GeoSet, self).__init__(content)

    @property
    def paths(self):
        return self._path

    @paths.setter
    def paths(self, paths):
        if not isinstance(paths, list):
            paths = [paths] * len(self)
        self._path = paths

    def __hash__(self):
        return hash(self.name + str(self))


def getGeoSet(mm_id, cls=None):
    """
    Find a GeoSet static attribute in the MeshedModelsIds class defined in config.py
    """
    if cls is None:
        import w7x

        cls = w7x.MeshedModelsIds
    for attr in cls.__dict__:
        thing = cls.__dict__[attr]
        if isinstance(thing, GeoSet):
            if mm_id in thing:
                return thing
    return None


def getGeoSets(mm_ids, cls=None):
    """

    Find the suiting GeoSet objects to the goeIds
    Examples:
        >>> import w7x
        >>> gss = w7x.geo.getGeoSets(list(range(320, 345)))
        >>> assert [gs.name for gs in gss] == ['baffle', 'tda', 'shield',
        ...                                    'panel', 'vessel']

    Raises:
        Not the complete GeoSet given
        >>> w7x.geo.getGeoSets(list(range(320, 321)))  # doctest: +ELLIPSIS
        Traceback (most recent call last):
            ...
        ValueError: GeoSet 'baffle' is not complete.

    """

    def checkCompleteness(geo_set, counter):
        if counter != len(geo_set) - 1:
            raise ValueError(
                "GeoSet '{geo_set.name}' is not complete.".format(**locals())
            )

    geoSets = []
    lastGeoSet = None
    counter = 0
    for i, mm_id in enumerate(mm_ids):
        geo_set = getGeoSet(mm_id, cls=cls)
        if geo_set is None:
            raise ValueError(
                "No GeoSet found for mm_id {mm_id} in cls '{cls}'".format(**locals())
            )
        if lastGeoSet == geo_set:
            counter += 1
            if geo_set[counter] != mm_id:
                raise ValueError(
                    "mm_id {mm_id} found at wrong position".format(**locals())
                )
        else:
            if i != 0:
                checkCompleteness(lastGeoSet, counter)
            counter = 0
            lastGeoSet = geo_set
            geoSets.append(geo_set)
    else:
        checkCompleteness(lastGeoSet, counter)

    return geoSets


if __name__ == "__main__":
    import doctest

    doctest.testmod()



"""
Transcoding input.txt file.

Yet reading only
"""
import transcoding as tc
import logging


#  TODO(@dboe): Reading only
#  TODO(@amerlo): Make main() vmec_id works
#  TODO(@dboe): Make read function independent of blocks position in input list
def get_transcoding():
    def f_read_number_string(number_string, name, dtype=float):
        """
        Read lines like "abc = 1 1 2 3 5 8"
        """
        try:
            return {name: float(number_string)}
        except Exception:
            return None

    def f_read_equal(line, name, event=None, dtype=float):
        """
        Read lines like "abc = 42.0"
        """
        if event is not None:
            if not event(line):
                return None
        _, _, number_string = line.rpartition("=")
        return f_read_number_string(number_string, name, dtype=dtype)

    #  Number of Poloidal Grid Points
    b_mpol = tc.Block(
        tc.Pattern(
            "MPOL = {mpol:d}",
            read_method=tc.read_method(f_read_equal, "mpol", lambda x: "MPOL" in x),
        ),
        start=tc.Trigger(lambda x: "MPOL" in x),
    )

    #  NTHETA
    b_ntheta = tc.Block(
        tc.Pattern(
            "NTHETA = {ntheta:d}",
            read_method=tc.read_method(f_read_equal, "ntheta", lambda x: "NTHETA" in x),
        ),
        start=tc.Trigger(lambda x: "NTHETA" in x),
    )

    #  Tolerance
    b_tolerance = tc.List(
        tc.Pattern(
            "{forceToleranceLevels:f}",
            read_method=tc.read_method(f_read_number_string, "forceToleranceLevels"),
        ),
        separator=" ",
        prefix=tc.Margin.alternatives(" FTOL_ARRAY = ", "FTOL_ARRAY = "),
        rigid=False,
        start=tc.Trigger(lambda x: "FTOL_ARRAY" in x),
    )
    #  Phi edge
    b_phiEdge = tc.Block(
        tc.Pattern(
            "PHIEDGE = {maxToroidalMagneticFlux:f}",
            read_method=tc.read_method(
                f_read_equal, "maxToroidalMagneticFlux", lambda x: "PHIEDGE" in x
            ),
        ),
        start=tc.Trigger(lambda x: "PHIEDGE" in x),
    )

    #  Pressure Scale
    b_pressureScale = tc.Block(
        tc.Pattern(
            "PRES_SCALE = {pressureScale:f}",
            read_method=tc.read_method(
                f_read_equal, "pressureScale", lambda x: "PRES_SCALE" in x
            ),
        ),
        start=tc.Trigger(lambda x: "PRES_SCALE" in x),
        barrier=tc.Trigger(lambda x: "CURTOR" in x),
        default={"pressureScale": 1.0},
    )

    #  Total Toroidal Current
    b_curtor = tc.Block(
        tc.Pattern(
            "CURTOR = {totalToroidalCurrent:f}",
            read_method=tc.read_method(
                f_read_equal, "totalToroidalCurrent", lambda x: "CURTOR" in x
            ),
        ),
        start=tc.Trigger(lambda x: "CURTOR" in x),
    )

    #  Radial resolution
    t_pressureProfile = tc.Trigger(
        lambda x: "PMASS_TYPE" in x or "IPMASS" in x, delay=0
    )
    b_radial = tc.List(
        tc.Pattern(
            "{numGridPointsRadial:d}",
            read_method=tc.read_method(f_read_number_string, "numGridPointsRadial"),
        ),
        separator=" ",
        prefix=tc.Margin.alternatives(" NS_ARRAY = ", "NS_ARRAY = ", "NS_ARRAY =   "),
        start=tc.Trigger(lambda x: "NS_ARRAY" in x),
        rigid=False,
        barrier=t_pressureProfile,
    )
    #  Pressure Profile
    t_pressureProfileBarrier = tc.Trigger(
        lambda x: "AM = " in x or " AM = " in x or " AM_AUX_F = " in x, delay=0
    )

    def f_pressureProfileHead(line):
        if "PMASS_TYPE" in line:
            return {"pressureProfileProfileType": line.split("= ")[1].replace("'", "")}
        if "IPMASS" in line:
            lut = {
                0: "power_series",
                1: "two_Lorentz",
                11: "Akima_spline",
                13: "cubic_spline",
            }
            ip = int(line.split("=")[1])
            return {"pressureProfileProfileType": lut[ip]}
        return None

    b_pressureProfileHead = tc.Block(
        tc.Pattern(
            "PMASS_TYPE = '{pressureProfileProfileType}'",
            read_method=tc.read_method(f_pressureProfileHead),
        ),
        start=t_pressureProfile,
        barrier=t_pressureProfileBarrier,
    )
    b_pressureProfile = tc.List(
        tc.Pattern(
            "{pressureProfileCoefficients:f}",
            read_method=tc.read_method(
                f_read_number_string, "pressureProfileCoefficients"
            ),
        ),
        separator=" ",
        rigid=False,
        prefix=tc.Margin.alternatives(" AM = ", "AM = ", "AM_AUX_F = "),
    )

    #  Coil Currents
    b_coilCurrents = tc.List(
        tc.Pattern(
            "{coilCurrents:f}",
            read_method=tc.read_method(f_read_number_string, "coilCurrents"),
        ),
        separator=" ",
        prefix=tc.Margin.alternatives(" EXTCUR = ", "EXTCUR = "),
        start=tc.Trigger(lambda x: "EXTCUR" in x),
        default={"coilCurrents": []},
    )

    #  Number of Iterations
    b_niter = tc.Block(
        tc.Pattern(
            "NITER = {niter:d}",
            read_method=tc.read_method(f_read_equal, "niter", lambda x: "NITER" in x),
        ),
        start=tc.Trigger(lambda x: "NITER" in x),
    )

    #  Number of Steps
    b_nstep = tc.Block(
        tc.Pattern(
            "NSTEP = {nstep:d}",
            read_method=tc.read_method(f_read_equal, "nstep", lambda x: "NSTEP" in x),
        ),
        start=tc.Trigger(lambda x: "NSTEP" in x),
    )

    #  Current Profile
    t_rAxis_start = tc.Trigger(lambda x: "RAXIS" in x, delay=0)

    def f_currentProfileHead(line):
        if "PCURR_TYPE" in line:
            return {"currentProfileProfileType": line.split("= ")[1].replace("'", "")}
        if "IPCURR" in line:
            lut = {
                0: "power_series",
                1: "sum_atan",
                2: "sum_atan",
                11: "Akima_spline_l",
                12: "Akima_spline_lp",
                13: "cubic_spline_l",
                14: "cubic_spline_lp",
            }
            ip = int(line.split("=")[1])
            return {"currentProfileProfileType": lut[ip]}
        return None

    b_currentProfileHead = tc.Block(
        tc.Pattern(
            "PCURR_TYPE = '{currentProfileProfileType}'",
            read_method=tc.read_method(f_currentProfileHead),
        ),
        start=tc.Trigger(lambda x: "PCURR_TYPE" in x or "IPCURR" in x, delay=0),
        barrier=tc.Trigger(
            lambda x: "AC = " in x or " AC = " in x or " AC_AUX_F = " in x, delay=0
        ),
        default={"currentProfileProfileType": "power_series"},
    )
    b_currentProfile = tc.List(
        tc.Pattern(
            "{currentProfileCoefficients:f}",
            read_method=tc.read_method(
                f_read_number_string, "currentProfileCoefficients"
            ),
        ),
        separator=" ",
        prefix=tc.Margin.alternatives("AC = ", " AC = ", "AC_AUX_F = "),
        rigid=False,
        barrier=t_rAxis_start,
        default={"currentProfileCoefficients": []},
    )

    #  rAxis
    b_rAxis = tc.List(
        "{RAXIS:f}",
        separator=" ",
        prefix=tc.Margin.alternatives("{tmp:5} = ", " {tmp:5} = "),
        rigid=False,
        start=t_rAxis_start,
    )

    #  zAxis
    b_zAxis = tc.List(
        "{ZAXIS:f}",
        separator=" ",
        prefix=tc.Margin.alternatives("{tmp:5} = ", " {tmp:5} = "),
        rigid=False,
        start=tc.Trigger(lambda x: "ZAXIS" in x, delay=0),
    )

    #  Boundary
    def f_boundary(line):
        if "rbc" in line or "RBC" in line:
            splt = line.split(",")
            if "" in splt:
                splt.remove("")
            if len(splt) != 4:
                return None
            vals = {
                "n": int(splt[0].split("(")[1]),
                "m": int(splt[1].split(")")[0]),
                "rbc": float(splt[1].split("=")[1]),
                "nz": int(splt[2].split("(")[1]),
                "mz": int(splt[3].split(")")[0]),
                "zbs": float(splt[3].split("=")[1]),
            }
            return vals
        return None

    b_boundary = tc.Table(
        tc.Pattern(
            " RBC({n:d},{m:d}) = {rbc:g}  ZBS({nz:d},{mz:d}) = {zbs:g}",
            read_method=tc.read_method(f_boundary),
        ),
        stop=tc.Trigger(lambda x: "/" in x, delay=0),
    )

    class CorrectedTranscoding(tc.Transcoding):
        def read(self, inp):
            newInp = []
            for line in inp:
                if "ZAXIS" in line:
                    if "rbc" in line:
                        newInp.append(line.split("rbc")[0])
                        newInp.append("rbc" + line.split("rbc")[1])
                    else:
                        break
                newInp.append(line)
            else:
                inp = newInp
            d = super(CorrectedTranscoding, self).read(inp)
            if "pressureProfileProfileType" not in d:
                d["pressureProfileProfileType"] = "power_series"
            if "currentProfileProfileType" not in d:
                d["currentProfileProfileType"] = "power_series"
            return d

    transcoding = CorrectedTranscoding(
        b_coilCurrents,
        b_niter,
        b_nstep,
        b_tolerance,
        b_phiEdge,
        b_mpol,
        b_ntheta,
        b_radial,
        b_pressureProfileHead,
        b_pressureProfile,
        b_currentProfileHead,
        b_currentProfile,
        b_pressureScale,
        b_curtor,
        b_rAxis,
        b_zAxis,
        b_boundary,
    )
    return transcoding


if __name__ == "__main__":
    import w7x

    log = logging.getLogger()
    log.setLevel(1)
    run = w7x.vmec.Run("w7x_ref_9")
    run._parse_input()
    print(run._input)



import logging
import tfields
import numpy as np

import w7x
from w7x import core


class Points3D(core.BasePoints3D):
    ws_server = w7x.Server.addr_components_db


class Base(core.Base):
    """
    Field Line Server Base
    """

    ws_server = w7x.Server.addr_components_db


class Polygon(Base):
    """
        element.vertices gives you the three points numbers to a triangle.
        This is normally refered to as face
    """

    prop_defaults = {"vertices": None}


class MeshedModel(Base):
    """
    Args:
        multiple ways:
            vertices (list)
            faces (list)

            - or -

            group from ObjFile

            - or -

            tfields.Mesh3D object
    Attributes:
        nodes (Points3D): = vertices (coordinates) of the points.
        elements (list[Polygon]): = faces (always three indices of points for a
            triangle). Starting at 1 here
    Examples:
        use with Mesh3D as inp
        >>> from w7x.flt import MeshedModel
        >>> import tfields
        >>> m = tfields.Mesh3D([[1,2,3], [3,3,3], [0,0,0], [5,6,7]],
        ...                    faces=[[0, 1, 2], [1, 2, 3]])
        >>> mm = MeshedModel(m)

        Get the osa type, in this case for field line server
        >>> fls = mm.as_input()

        return Mesh3D works
        >>> bool((m == mm.as_mesh3d()).all())
        True

        create with meshed Model from fls works
        >>> m2 = MeshedModel(fls).as_mesh3d()
        >>> assert tfields.Points3D(m2).equal(
        ...     [[ 1.,  2.,  3.],
        ...      [ 3.,  3.,  3.],
        ...      [ 0.,  0.,  0.],
        ...      [ 5.,  6.,  7.]])
        >>> assert tfields.Tensors(m2.faces).equal([[0, 1, 2], [1, 2, 3]])

    """

    prop_defaults = {
        "nodes": None,
        "elements": None,
        "nodesIds": None,
        "elementsIds": None,
    }

    def __init__(self, *args, **kwargs):
        args = list(args)
        if len(args) > 1:
            logging.error(" Implementation did not work.")
            raise NotImplementedError(
                " Implementation with args %s not yet implemented!" % args
            )
        elif len(args) == 1 and issubclass(args[0].__class__, tfields.Mesh3D):
            mesh3D = args.pop(0)
            nodes = Points3D(mesh3D)
            faces = mesh3D.faces + 1
            kwargs["nodes"] = kwargs.pop("nodes", nodes)
            kwargs["elements"] = kwargs.pop(
                "elements", [Polygon(vertices=face) for face in faces]
            )
        super(MeshedModel, self).__init__(*args, **kwargs)

    @classmethod
    def from_mm_id(cls, mm_id):
        compDBServer = w7x.get_server(w7x.Server.addr_components_db)
        meshedModelFLSType = compDBServer.service.getComponentData(mm_id)[0]
        return cls(meshedModelFLSType)

    def as_mesh3d(self):
        faces = np.array([pol.vertices for pol in self.elements])
        faces -= 1
        return tfields.Mesh3D(Points3D(self.nodes), faces=faces)


class HistoryNote(Base):
    prop_defaults = {
        "comment": None,  # str
        "author": None,  # str
        "date": None,  # datetime
        "method": None,  # str
    }


class ComponentStorageInfo(Base):
    prop_defaults = {
        "comment": None,  # str
        "polygonTypes": None,  # List[int]
        "numNodes": None,  # int
        "id": None,  # str
        "machine": "w7x",  # str, ws default is None
        "min": None,  # List[float]
        "numElements": None,  # int
        "state": None,  # str
        "location": None,  # str
        "accuracy": None,  # float
        "max": None,  # List[float]
        "name": None,  # str
        "subids": None,  # List[str]
        "author": None,  # str
        "method": None,  # str
        "resolution": None,  # float
        "hisotry": None,  # List(HistoryNote)
    }


class FullComponentEntry(Base):
    prop_defaults = {
        "info": None,  # ComponentStorageInfo
        "data": None,  # MeshedModel
    }

    @property
    def parent(self):
        return self._parent



"""
Transcoding threed1.txt file.

Yet reading only
"""
import transcoding as tc


#  TODO(@dboe): Output dict get redundant keys and values. Run main() to
#               reproduce
def get_transcoding():

    #  Header
    b_header = tc.Block(
        " THIS IS {edition}, VERSION {version:f}",
        " Lambda: Full Radial Mesh. L-Force: hybrid full/half.",
        "",
        " COMPUTER: {computer}   OS: {os}   RELEASE: {release}  DATE = {date} TIME = {time}",
        " SHOT ID.: {shotId}  SEQ. NO.: {seqNo}",
        " TIME SLICE = {timeSlize} ms",
    )

    #  Pressure Profile Header
    t_pressureProfile = tc.Trigger(lambda x: " MASS PROFILE COEFFICIENTS" in x, delay=0)
    b_pressureProfileHead = tc.Block(
        " MASS PROFILE COEFFICIENTS - {massProfileCoeffsExplanation}:",
        " PMASS parameterization type is '{paraType}'",
        " -----------------------------------",
        start=t_pressureProfile,
    )

    #  Pressure Profile Coefficients
    b_pressureProfile = tc.List(
        "{pressureProfileCoefficients:e}", separator="   ", rigid=False, prefix="  "
    )

    #  Force Iterations
    b_forceIterHead = tc.Block(
        " BEGIN FORCE ITERATIONS",
        " -----------------------",
        "",
        "",
        start=tc.Trigger(lambda x: " BEGIN FORCE ITERATIONS" in x, delay=0),
    )

    #  Fourier Modes
    b_forceIterSectionHead = tc.Block(
        tc.Pattern(
            "  NS = {ns:4d} NO. FOURIER MODES = {nFourierModes:4d} FTOLV = "
            "{ftol:6.3e} NITER = {niter:6d}",
            read_method=tc.Pattern(
                "  NS = {ns:4d} NO. FOURIER MODES = {nFourierModes:4d}"
                " FTOLV = {ftol:.3e}"
            ),
        ),
        tc.Pattern(
            "  PROCESSOR COUNT - RADIAL: {processorCountRadial:4d}  "
            "VACUUM: {processorCountVacuum:4d}",
            read_method=tc.Pattern(
                "  PROCESSOR COUNT - RADIAL: {processorCountRadial:4d}  "
                "VACUUM: {processorCountVacuum:4d}"
            ),
        ),
        start=tc.Trigger(lambda x: "NO. FOURIER MODES" in x, delay=0),
    )

    #  Fourier Modes Table
    #  TODO(@amerlo): Case of power >= 100, pattern breaks
    t_forceIterTableStart = tc.Trigger(lambda x: "ITER    FSQR      FSQZ" in x, delay=1)
    t_forceIterTail = tc.Trigger(lambda x: "d(ln W)/dt" in x, delay=0)
    b_forceIterTable = tc.Table(
        "{iter:6}  "
        "{FSQR:.2e}  {FSQZ:.2e}  {FSQL:.2e}  "
        "{fsqr:.2e}  {fsqz:.2e}  {fsql:.2e}  "
        "{DELT:.2e}  {RAX:.3e}  {FHMD:.4e}  "
        "{BETA:.3e}  {M:.3f} {DELBSQ:.2e} "
        "{FEDGE:.2e}",
        start=t_forceIterTableStart,
        stop=t_forceIterTail,
    )

    #  MHD Energy
    b_forceIterTail = tc.Block(
        " MHD Energy = {MHDEnergy:.6e}   d(ln W)/dt = {dlnWdt:.3e}   d(ln R0)/dt = {dlnRdt:.3e}",
        "",
    )

    t_profileHeadStart = tc.Trigger(lambda x: "JSUPU" in x and "BSUBU" in x, delay=0)

    #  Fourier Modes Loop
    l_forceIterLoop = tc.Loop(
        "forceIter",
        [b_forceIterSectionHead, b_forceIterTable, b_forceIterTail],
        stop_iter=t_profileHeadStart,
        head=b_forceIterHead,
    )

    #  Current Profile
    b_profileHead = tc.Block(
        "      S     <RADIAL    TOROIDAL      IOTA      <JSUPU>    <JSUPV>     d(VOL)/   d(PRES)/"
        "    <M>     PRESF    <BSUBU>    <BSUBV>      <J.B>      <B.B>",
        "             FORCE>      FLUX                                         d(PHI)     d(PHI)"
        "                             ",
        "-------------------------------------------------------------------------------------"
        "---------------------------------------------------------------",
        "",
        start=t_profileHeadStart,
    )

    t_profileTableStop = tc.Trigger(lambda x: len(x) == 0, delay=0)
    b_profileTable = tc.Table(
        " {S:.2e} {RADIAL_FORCE:.2e} {TOROIDAL_FLUX:.4e} {IOTA:.4e} {JSUPU:.3e} {JSUPV:.3e} "
        "{dVOLdPhi:.3e} {dPRESdPHI:.3e} {M:.3f} {PRESF:.3e} {BSUBU:.3e} {BSUBV:.3e} "
        "{JB:.3e} {BB:.3e}",
        stop=t_profileTableStop,
    )

    #  Geometric and Magnetic Quantities
    t_geoMagnetic = tc.Trigger(
        lambda x: "Geometric and Magnetic Quantities" in x, delay=0
    )
    b_geoMagnetic = tc.Block(
        " Geometric and Magnetic Quantities",
        " -----------------------------------------------------------------------",
        " Aspect Ratio          = {AspectRatio:14.5f}",
        " Mean Elongation       = {MeanElongation:14.5f}",
        " Plasma Volume         = {PlasmaVolume:14.5f} [M**3]",
        " Cross Sectional Area  = {CrossSectionalArea:14.5f} [M**2]",
        " Normal Surface Area   = {NormalSurfaceArea:14.5f} [M**2]",
        " Poloidal Circumference= {PoloidalCircumference:14.5f} [M]",
        " Major Radius          = {MajorRadius:14.5f} [M]"
        " (from Volume and Cross Section)",
        " Minor Radius          = {MinorRadius:14.5f} [M] (from Cross Section)",
        " Minimum (inboard)  R  = {MinimumInboardR:14.5f} [M]",
        " Maximum (outboard) R  = {MinimumOutboardR:14.5f} [M]",
        " Maximum height     Z  = {MaximumHeightZ:14.5f} [M]",
        " Waist (v = 0)   in R  = {WaistZeroR:14.5f} [M]",
        " Full Height(v = 0)    = {FullHeightZero:14.5f} [M]",
        " Waist (v = pi)  in R  = {WaistPiR:14.5f} [M]",
        " Full Height(v = pi)   = {FullHeightPi:14.5f} [M]",
        " Toroidal Flux         = {ToroidalFlux:14.5f} [Wb]",
        " Toroidal Current      = {ToroidalCurrent:14.5f} [MA]",
        " RBtor(s=1)            = {RBtorZero:14.5f} [T-m]",
        " RBtor(s=0)            = {RBtorOne:14.5f} [T-m]",
        " Volume Average B      = {VolumeAverageB:14.5f} [T]",
        " Ion Larmor Radius     = {IonLarmorRadius:14.5f} [M] X Ti(keV)**0.5",
        " <J||**2>/<J-perp**2>  = {JJPerp:14.5f} (Vol. Averaged)",
        " <JPS**2>/<J-perp**2>  = {JPSJPerp:14.5f} (Vol. Averaged)",
        start=t_geoMagnetic,
    )

    #  Magnetic Fields and Pressure
    b_magField = tc.Block(
        " Magnetic Fields and Pressure",
        " -----------------------------------------------------------------------",
        " Volume Integrals (Joules) and Volume Averages (Pascals)",
        "                        Integral      Average",
        " pressure         =   {PressureIntegral:.6e}  {PressureAverage:.6e}",
        " bpol**2 /(2 mu0) =   {PoloidalMagneticEnergyIntegral:.6e}  {PoloidalMagneticEnergyAverage:.6e}",  # noqa
        " btor**2/(2 mu0)  =   {ToroidalMagneticEnergyIntegral:.6e}  {ToroidalMagneticEnergyAverage:.6e}",  # noqa
        " b**2/(2 mu0)     =   {MagneticEnergyIntegral:.6e}  {MagneticEnergyAverage:.6e}",
        " EKIN (3/2p)      =   {KineticEnergyIntegral:.6e}  {KineticEnergyAverage:.6e}",
        start=tc.Trigger(lambda x: "Magnetic Fields and Pressure" in x, delay=0),
    )

    #  Magnetic Axis Coefficients
    t_axis_start = tc.Trigger(lambda x: "MAGNETIC AXIS COEFFICIENTS" in x, delay=4)
    t_axis_stop = tc.Trigger(lambda x: not x, delay=0, skip=1)
    b_axis = tc.Table(
        " {na:4d} {rac:11.4e} {zas:11.4e}", start=t_axis_start, stop=t_axis_stop
    )

    t_boundary_start = tc.Trigger(
        lambda x: "nb  mb      rbc         zbs      vacpot_s  |B|_c(s=.5) |B|_c(s=1.)"
        in x,
        delay=2,
    )
    t_boundary_stop = tc.Trigger(lambda x: not x, delay=0, skip=1)
    b_boundary = tc.Table(
        " {nb:3d} {mb:3d} {rbc:11.4e} {zbs:11.4e}"
        " {vacpot_s:11.4e} {B_c_s_half:11.4e}"
        " {B_c_s_tenth:11.4e}",
        start=t_boundary_start,
        stop=t_boundary_stop,
    )

    #  VMEC file
    b_file = tc.Block(" FILE : {vmecFile}", start=tc.Trigger(lambda x: "FILE :" in x))

    #  Computational time
    b_time = tc.Block(
        "    TOTAL COMPUTATIONAL TIME (SEC) {totalTime:12.2f}",
        start=tc.Trigger(lambda x: "TOTAL COMPUTATIONAL" in x),
    )

    inst = tc.Transcoding(
        b_header,
        b_pressureProfileHead,
        b_pressureProfile,
        l_forceIterLoop,
        b_profileHead,
        b_profileTable,
        b_geoMagnetic,
        b_magField,
        b_axis,
        b_boundary,
        b_file,
        b_time,
    )
    return inst


if __name__ == "__main__":
    import w7x
    import logging

    log = logging.getLogger()
    log.setLevel(1)
    run = w7x.vmec.Run(
        "w7x_v_0.1.0.dev8_id_1000_1000_1000_1000_+0000_-0250_pres_00_it_9"
    )
    run._parse_threed1()
    print(run.beta())



"""
specific methods for poincare plotting
"""
import tfields
import rna


def plot_poincare_surfaces(poincareSurfaces, **kwargs):
    """
    Args:
        poincareSurfaces (list of Points3D): each Points3D instance is one
            fieldLine followed around the torus
    """
    method = kwargs.pop('method', 'scatter')
    po = rna.plotting.PlotManager(kwargs)
    rMin = po.pop('rMin', 4.0)
    rMax = po.pop('rMax', 6.6)
    zMin = po.pop('zMin', -1.3)
    zMax = po.pop('zMax', +1.3)
    phiRad = po.pop('phiRad', None)

    po.setdefault('y_index', 2)
    po.setdefault('labels', ['r (m)', r'$\phi$ (rad)', 'z (m)'])
    if method == 'scatter':
        po.setdefault('marker', '.')
        po.setdefault('s', 1)
    po.setdefault('method', method)
    color_given = True
    if 'color' not in po.plot_kwargs:
        color_given = False
        cmap, _, _ = po.get_norm_args()
        color_cycle = rna.plotting.colors.color_cycle(cmap,
                                                      len(poincareSurfaces))
    elif isinstance(po.get('color'), list):
        color_given = False  # hack to set the color from list
        color_cycle = iter(po.get('color'))
    artists = []
    for i, surfacePoints in enumerate(poincareSurfaces):
        with surfacePoints.tmp_transform(tfields.bases.CYLINDER):
            phiSurface = surfacePoints[:, 1]
            if phiRad is None:
                phiRad = phiSurface[0]
            if bool((phiSurface != phiRad).any()):
                continue
            if not color_given:
                po.set('color', next(color_cycle))
            artists.append(surfacePoints.plot(axes=po.axes, **po.plot_kwargs))
    rna.plotting.set_aspect_equal(po.axes)
    po.axes.set_xlim(rMin, rMax)
    po.axes.set_ylim(zMin, zMax)
    return artists


def plot_poincare_geometries(geometries, **kwargs):
    po = rna.plotting.PlotManager(kwargs)
    po.setdefault('method', 'plot')
    po.setdefault('lw', 1)
    artists = []
    for p in range(len(geometries)):
        for g in range(len(geometries[p])):
            artists.extend(plot_poincare_surfaces(geometries[p][g],
                                                  axes=po.axes,
                                                  **po.plot_kwargs))
    return artists



"""Top-level package of w7x."""

__author__ = """Daniel Böckenhoff"""
__email__ = "dboe@ipp.mpg.de"
__version__ = "0.3.0"

from .geo import GeoSet, getGeoSet  # NOQA
from .config import *  # NOQA
from .core import is_w7x_instance, get_server, get_ws_class, run_service  # NOQA
from . import transcodings  # NOQA
from . import flt  # NOQA
from .flt import MagneticConfig, Machine, Points3D  # NOQA
from . import vmec  # NOQA
from . import extender  # NOQA
from . import plotting  # NOQA
from .plotting.poincare import plot_poincare_surfaces, plot_poincare_geometries  # NOQA
from .plotting.tile_loads import plot_tile_loads  # NOQA
from .plotting.vmec import plot_fourier  # NOQA
from . import batch  # NOQA


def switch_offline():
    """Allows loading old results without calling the web-service."""
    from . import core

    core.get_ws_class = lambda x, y: type(None)
